/** @file ag_gui_base.cpp

	Skelton for retropc emulator
	SDL + Ager edition

	@author Sasaji
	@date   2012.3.2

	@brief [ ag_gui_base ]
*/

#include "ag_gui_base.h"
#include "../emu.h"
#include "ag_volume_dlg.h"
#include "ag_file_dlg.h"

AG_GUI_BASE::AG_GUI_BASE(EMU *parent_emu)
{
	emu = parent_emu;

	memset(drivers, 0, sizeof(drivers));
	now_driver = NULL;

	exit_event = 0;
	first = 120;

	filebox = NULL;
	volumebox = NULL;

	now_showing = false;
}

AG_GUI_BASE::~AG_GUI_BASE()
{
	// save AGAR config
	ConfigSave();

	delete volumebox;
	delete filebox;

	AG_DestroyGUI();
	AG_DestroyGUIGlobals();
	AG_Destroy();
}

int AG_GUI_BASE::InitCore()
{
	int rc = 0;
	Uint32 flags = 0;

#ifdef _DEBUG
	flags |= AG_VERBOSE;
#endif

	// init AGAR GUI
	if ((rc = AG_InitCore(CONFIG_NAME, flags)) == -1) {
		emu->out_log(LOG_ERROR, _T("AG_InitCore: %s"), AG_GetError());
	}
	// load AGAR config
	ConfigLoad();

	// dialog
	filebox = new AG_FILE_DLG(emu, this);
	volumebox = new AG_VOLUME_DLG(emu, this);

	return rc;
}

int AG_GUI_BASE::InitVideoSDL(SDL_Surface *screen, int width, int height)
{
	int rc = 0;

	if (screen->flags & SDL_OPENGL) {
		now_driver = &drivers[1];
	} else {
		now_driver = &drivers[0];
	}

	if (now_driver->win == NULL) {
#if 0
		if ((rc = AG_InitGraphics(NULL)) == -1) {
			emu->out_log(LOG_ERROR, "AG_InitGraphics: %s", AG_GetError());
			return rc;
		}
		// Display main window
		drv->win = AG_WindowNewNamedS(AG_WINDOW_NOMAXIMIZE, "test gui");
		if (drv->win == NULL) {
			rc = -1;
			emu->out_log(LOG_ERROR, "AG_WindowNew: %s", AG_GetError());
			return rc;
		}
#else
		// Initialize Agar-GUI to reuse display
		if ((rc = AG_InitVideoSDL(screen, AG_VIDEO_NOFRAME | AG_VIDEO_OVERLAY)) == -1) {
			emu->out_log(LOG_ERROR, _T("AG_InitVideoSDL: %s"), AG_GetError());
			return rc;
		}
		// attach ag window to SDL
		now_driver->win = AG_WindowNew(AG_WINDOW_PLAIN | AG_WINDOW_NOBACKGROUND | AG_WINDOW_NOBUTTONS);
		if (now_driver->win == NULL) {
			rc = -1;
			emu->out_log(LOG_ERROR, _T("AG_WindowNew: %s"), AG_GetError());
			return rc;
		}
#endif
		now_driver->ops = agDriverOps;
		now_driver->sw  = agDriverSw;

		// create menu
		if ((rc = CreateMenu()) == -1) {
			return rc;
		}
		// set user event
		AG_AddEvent(now_driver->win, "user-process", OnProcessUserEvent, "%Cp", this);

	} else {
		// change driver
		agDriverOps = now_driver->ops;
		agDriverSw = now_driver->sw;

		AG_SetVideoSurfaceSDL(screen);
		AG_ResizeDisplay(width, height);
	}
	AG_WindowSetGeometry(now_driver->win, 0, 0, width, height);

	return rc;
}

int AG_GUI_BASE::ResizeDisplay(int width, int height)
{
	return AG_ResizeDisplay(width, height);
}

int AG_GUI_BASE::CreateMenu()
{
	int rc = 0;

	if (now_driver == NULL) {
		rc = -1;
		return rc;
	}

	now_driver->menu = AG_MenuNew(now_driver->win, 0);
	if (now_driver->menu == NULL) {
		rc = -1;
		emu->out_log(LOG_ERROR, _T("Error: AG_MenuNew: %s"), AG_GetError());
		return rc;
	}

	set_menu_item(now_driver->menu);

	Show(true);

	return rc;
}

int AG_GUI_BASE::CreateGlobalKeys()
{
	int rc = 0;

	set_global_keys();

	return rc;
}

/**
 * global keys
 */
bool AG_GUI_BASE::ExecuteGlobalKeys(int key, Uint32 mod)
{
	return false;
}

/**
 * user event
 */
void AG_GUI_BASE::ProcessUserEvent(int id)
{
}

AG_Window *AG_GUI_BASE::GetWindow()
{
	return now_driver->win;
}

AG_DLG *AG_GUI_BASE::GetDlgPtr(int)
{
	return NULL;
}

void AG_GUI_BASE::Show(bool val)
{
	if (val) {
		AG_WindowShow(now_driver->win);
	} else {
		AG_WindowHide(now_driver->win);
	}
	now_showing = val;
}

void AG_GUI_BASE::Pause(bool val)
{
	emu->set_pause(1, val);
	emu->mute_sound(val);
}

/**
 * gui event loop (AGAR)
 *
 * @note
 * This method no longer use.
 */
int AG_GUI_BASE::EventLoop()
{
	AG_Window *win;
	Uint32 t1, t2;
	AG_DriverEvent dev;

	AG_WindowShow(now_driver->win);

	exit_event = 0;
	t1 = AG_GetTicks();
	while(exit_event == 0) {
		t2 = AG_GetTicks();
		if (t2 - t1 > 16) {
			/*
			 * Case 1: Update the video display.
			 */
			emu->mix_screen();

			AG_LockVFS(&agDrivers);
			if (agDriverSw) {
				AG_BeginRendering(agDriverSw);
				AG_FOREACH_WINDOW(win, agDriverSw) {
					if (!win->visible) {
						continue;
					}
					AG_ObjectLock(win);
					AG_WindowDraw(win);
					AG_ObjectUnlock(win);
				}
				AG_EndRendering(agDriverSw);
			}
			AG_UnlockVFS(&agDrivers);
			t1 = AG_GetTicks();

		} else if (AG_PendingEvents(NULL) > 0) {
			/*
			 * Case 2: There are events waiting to be processed.
			 */
			do {
//				if (AG_GetNextEvent(NULL, &dev) == 1) {
				if (GetNextEvent(agDriverSw, &dev) == 1) {
					event_proc(&dev);
					AG_ProcessEvent(NULL, &dev);
				}
			} while (AG_PendingEvents(NULL) > 0);
			t1 = t2 - 17;

		} else if (AG_TIMEOUTS_QUEUED()) {
			/*
			 * Case 3: There are AG_Timeout(3) callbacks to run.
			 */
			AG_ProcessTimeouts(t2);

		} else {
			/*
			 * Case 4: Nothing to do, idle.
			 */
			AG_Delay(1);

		}
	}
	return exit_event;
}

/**
 * gui get event (AGAR)
 */
int AG_GUI_BASE::GetNextEvent(void *obj, AG_DriverEvent *dev)
{
	AG_Driver *drv = (AG_Driver *)obj;

	if (SDL_PollEvent(&sdl_event) == 0) {
		return (0);
	}
	if (sdl_event.type == SDL_KEYDOWN || sdl_event.type == SDL_KEYUP) {
		translate_keycode(&sdl_event.key);
	}
	AG_SDL_TranslateEvent(drv, &sdl_event, dev);
	return (1);
}

/**
 * gui process event (AGAR)
 */
int AG_GUI_BASE::ProcessEvent(SDL_Event *event)
{
	AG_DriverEvent dev;

	sdl_event = *event;
	exit_event = 0;
	if (first > 0) first--;

	if (sdl_event.type == SDL_KEYDOWN || sdl_event.type == SDL_KEYUP) {
		translate_keycode(&sdl_event.key);
	}
	AG_SDL_TranslateEvent((void *)agDriverSw, &sdl_event, &dev);
	event_proc(&dev);
	AG_ProcessEvent(NULL, &dev);
	return (exit_event);
}

/**
 * gui mix screen (AGAR)
 */
int AG_GUI_BASE::MixGuiScreen()
{
	AG_Window *win;

	if (!now_showing) return 0;

	AG_LockVFS(&agDrivers);
	if (agDriverSw) {
		AG_BeginRendering(agDriverSw);
		AG_FOREACH_WINDOW(win, agDriverSw) {
			if (!win->visible) {
				continue;
			}
			AG_ObjectLock(win);
			AG_WindowDraw(win);
			AG_ObjectUnlock(win);
		}
		AG_EndRendering(agDriverSw);
	}
	AG_UnlockVFS(&agDrivers);

	return 1;
}

/**
 * process timeout (AGAR)
 */
int AG_GUI_BASE::ProcessTimeout()
{
	Uint32 t1 = AG_GetTicks();
	if (AG_TIMEOUTS_QUEUED()) {
		// There are AG_Timeout(3) callbacks to run.
		AG_ProcessTimeouts(t1);
		return (1);
	}
	return (0);
}

/**
 *	gui event procedure (AGAR)
 */
void AG_GUI_BASE::event_proc(AG_DriverEvent *dev)
{
	switch(dev->type) {
	case AG_DRIVER_MOUSE_MOTION:		// Cursor moved
		gui_focus_out(dev->data.motion.x, dev->data.motion.y);
		break;
	case AG_DRIVER_MOUSE_BUTTON_DOWN:	// Mouse button pressed
		break;
	case AG_DRIVER_MOUSE_BUTTON_UP:		// Mouse button released
		gui_focus_out(dev->data.motion.x, dev->data.button.y);
		break;
	case AG_DRIVER_MOUSE_ENTER:			// Mouse entering window (MW)
		break;
	case AG_DRIVER_MOUSE_LEAVE:			// Mouse leaving window (MW)
		break;
	case AG_DRIVER_FOCUS_IN:			// Focus on window (MW)
		break;
	case AG_DRIVER_FOCUS_OUT:			// Focus out of window (MW)
		break;
	case AG_DRIVER_KEY_DOWN:			// Key pressed
		break;
	case AG_DRIVER_KEY_UP:				// Key released
		break;
	case AG_DRIVER_EXPOSE:				// Video update needed
		break;
	case AG_DRIVER_VIDEORESIZE:			// Video resize request
		break;
	case AG_DRIVER_CLOSE:				// Window close request
		emu->out_debug(_T("AG_DRIVER_CLOSE"));
		exit_event = -1;
		break;
	case AG_DRIVER_UNKNOWN:
		break;
	}
}

void AG_GUI_BASE::gui_focus_out(int x, int y)
{
	if (now_driver->menu->selecting == 0 && (y <= 0 || 24 <= y || x <= 0 || SCREEN_WIDTH <= x)) {
		AG_Window *win;
		int i = 0;
		// count active widgets
		AG_FOREACH_WINDOW(win, agDriverSw) {
			if (!win->visible) {
				continue;
			}
			i++;
		}
		if (i == 1 && first <= 0) {
			// menu only
//			emu->out_debug(_T("AG_DRIVER_MOUSE_MOTION: y:%d -> Go SDL "),y);
			exit_event = 1;
			AG_WindowHide(now_driver->win);
		}
	}
}

/**
 * create menu
 */
void AG_GUI_BASE::set_menu_item(AG_Menu *menu)
{
	AG_MenuItem *mt_control;

	mt_control = AG_MenuNode(menu->root, "Control", NULL);
	{
		AG_MenuAction(mt_control, "  Quit", NULL, OnSelectExit, "%Cp", this);
	}
}

/**
 * create global keys
 */
void AG_GUI_BASE::set_global_keys()
{
// oops....Why cannot you set the args for event structure such like AG_SetEvent???
//	AG_BindGlobalKeyEv(AG_KEY_P, AG_KEY_LALT, OnSelectPause);
}

/****************************************
 *	gui translate keycode
 */
void AG_GUI_BASE::translate_keycode(SDL_KeyboardEvent *e)
{
#if defined(_WIN32)
	switch(e->keysym.scancode) {
		case 0x73:	// underscore
			e->keysym.sym = SDLK_UNDERSCORE;
			break;
	}
#elif defined(__APPLE__) && defined(__MACH__)
	switch(e->keysym.scancode) {
		case 0x5d:      // yen
			e->keysym.sym = SDLK_BACKSLASH;
			break;
		case 0x5e:      // underscore
			e->keysym.sym = SDLK_UNDERSCORE;
			break;
	}
#endif
}

/****************************************
 * Event Operation
 */
///
void AG_GUI_BASE::update_multi_volume_list(AG_MenuItem *mi, int drv)
{
	AG_MenuItem *ms;
	char name[32];
	bool flag = false;
	bool upd = false;	

	upd = emu->changed_cur_bank(drv);

	if (!upd) return;

	AG_MenuItemFreeChildren(mi);

	d88_file_t *d88_file = emu->get_d88_bank(drv);

	for(int num=0; num<d88_file->bank_nums; num++) {
		sprintf(name, "%c %2d: %s"
			, num == d88_file->cur_bank ? '*' : ' '
			, num+1
			, d88_file->bank[num].name[0] != _T('\0') ? d88_file->bank[num].name : "(no label)");
		ms = AG_MenuAction(mi, name, NULL, OnSelectVolumeFloppyDisk, "%Cp %i %i", this, drv, num);
		AG_GUI_MENU_ENABLE(ms, true);
		flag = true;
	}
	if (d88_file->bank_nums == 1) {
		AG_GUI_MENU_ENABLE(ms, false);
	}
	if (!flag) {
		ms = AG_MenuNode(mi, "None", NULL);
		AG_GUI_MENU_ENABLE(ms, false);
	}
}

///
void AG_GUI_BASE::update_recent_list(AG_MenuItem *mi, int type, int drv)
{
	AG_MenuItem *ms;
	char path[_MAX_PATH+8];
	char *file;
	bool flag = false;
	bool upd = false;

	switch(type) {
	case AG_FILE_DLG::DATAREC:
		upd = config.updated_recent_datarec_path;
		config.updated_recent_datarec_path = false;
		break;
	case AG_FILE_DLG::FLOPPY:
		upd = config.updated_recent_disk_path[drv];
		config.updated_recent_disk_path[drv] = false;
		break;
	}
	if (!upd) return;

	AG_MenuItemFreeChildren(mi);

	for(int num=0; num<MAX_HISTORY; num++) {
		switch(type) {
		case AG_FILE_DLG::DATAREC:
			file = config.recent_datarec_path[num].path;
			if (file[0] == '\0') break;
			strcpy(path, UTILITY::conv_from_native_path(file, _MAX_PATH));
			strcpy(path, UTILITY::trim_center(path, 56));
			if (config.recent_datarec_path[num].num > 0) {
				sprintf(&path[strlen(path)], " : %d", config.recent_datarec_path[num].num + 1);
			}
			ms = AG_MenuAction(mi, path, NULL, OnSelectRecentDataRec, "%Cp %s", this, file);
			AG_GUI_MENU_ENABLE(ms, true);
			flag = true;
			break;
		case AG_FILE_DLG::FLOPPY:
			file = config.recent_disk_path[drv][num].path;
			if (file[0] == '\0') break;
			strcpy(path, UTILITY::conv_from_native_path(file, _MAX_PATH));
			strcpy(path, UTILITY::trim_center(path, 56));
			if (config.recent_disk_path[drv][num].num > 0) {
				sprintf(&path[strlen(path)], " : %d", config.recent_disk_path[drv][num].num + 1);
			}
			ms = AG_MenuAction(mi, path, NULL, OnSelectRecentFloppyDisk, "%Cp %i %s", this, drv, file);
			AG_GUI_MENU_ENABLE(ms, true);
			flag = true;
			break;
		}
	}
	if (!flag) {
		ms = AG_MenuNode(mi, "None", NULL);
		AG_GUI_MENU_ENABLE(ms, false);
	}
}
// Exit
void AG_GUI_BASE::Exit(void)
{
	AG_QuitGUI();
}
////////////////////////////////////////
// Reset
void AG_GUI_BASE::Reset(void)
{
	emu->reset();

//	rec_next_time = rec_accum_time = 0;
}
// SpecialReset
void AG_GUI_BASE::SpecialReset(void)
{
	emu->special_reset();

//	rec_next_time = rec_accum_time = 0;
}
// WarmReset
void AG_GUI_BASE::WarmReset(void)
{
	emu->warm_reset();
}
// change dipswitch
void AG_GUI_BASE::Dipswitch(int bit)
{
	emu->change_dipswitch(bit);
}
// change pause
void AG_GUI_BASE::Pause(void)
{
	bool vm_pause = emu->get_pause(2);
	vm_pause = (vm_pause ? false : true);
	emu->set_pause(2, vm_pause);
}
// change cpu power
void AG_GUI_BASE::CPUPower(int num)
{
	emu->change_cpu_power(num);
}

// Change Window Mode
void AG_GUI_BASE::ChangeWindowMode(int num)
{
	emu->change_screen_mode(num);
}
// Change Screen Mode
void AG_GUI_BASE::ChangeScreenMode(int num)
{
	emu->change_screen_mode(num + 8);
}
// Toggle Scan Line
void AG_GUI_BASE::ToggleScanLine(void)
{
	emu->change_screen_scanline();
}
// create Volume dialog
void AG_GUI_BASE::CreateVolumeBox(void)
{
	AG_WindowShow(now_driver->win);
	volumebox->Create();
}
// show Data Recorder Load Dialog
void AG_GUI_BASE::ShowLoadDataRecDialog(void)
{
	AG_WindowShow(now_driver->win);
	filebox->CreateLoad(AG_FILE_DLG::DATAREC, "Play Data Recorder Tape"
		, "*.tap", config.initial_datarec_path, NULL);
}
// Load Data Recorder
bool AG_GUI_BASE::LoadDataRec(char *file_path)
{
	return emu->play_datarec(file_path);
}
// show Data Recorder Save Dialog
void AG_GUI_BASE::ShowSaveDataRecDialog(void)
{
	AG_WindowShow(now_driver->win);
	filebox->CreateSave(AG_FILE_DLG::DATAREC, "Record Data Recorder Tape"
		, "*.tap", config.initial_datarec_path, NULL);
}
// Save Data Recorder
bool AG_GUI_BASE::SaveDataRec(char *file_path)
{
	return emu->rec_datarec(file_path);
}
// show Floppy Disk Open Dialog
void AG_GUI_BASE::ShowOpenFloppyDiskDialog(int drv)
{
	char title[128];

	sprintf(title, "Open Floppy Disk %d", drv);
	AG_WindowShow(now_driver->win);
	filebox->CreateLoad(AG_FILE_DLG::FLOPPY, title
		, "*.d88", config.initial_disk_path, NULL, (void *)drv);
}
// open Floppy Disk 
bool AG_GUI_BASE::OpenFloppyDisk(int drv, _TCHAR* path, int bank_num, uint32 flags, bool multiopen)
{
	return emu->open_disk_by_bank_num(drv, path, bank_num, flags, multiopen);
}
// open Floppy Disk 
bool AG_GUI_BASE::OpenFloppyDiskSelectedVolume(int drv, int bank_num)
{
	return emu->open_disk_with_sel_bank(drv, bank_num);
}
// close Floppy Disk
void AG_GUI_BASE::CloseFloppyDisk(int drv)
{
	emu->close_disk(drv);
}
void AG_GUI_BASE::LoadStatus(char *file_path)
{
	emu->load_state(file_path);
}

void AG_GUI_BASE::SaveStatus(char *file_path)
{
	emu->save_state(file_path);
}

void AG_GUI_BASE::LoadAutoKey(char *file_path)
{
	emu->start_auto_key(file_path);
}

void AG_GUI_BASE::LoadRecKey(char *file_path)
{
	emu->play_reckey(file_path);
}

void AG_GUI_BASE::SaveRecKey(char *file_path)
{
	emu->record_reckey(file_path);
}

void AG_GUI_BASE::SavePrinter(int drv, char *file_path)
{
	emu->save_printer(drv, file_path);
}

/****************************************
 * Event Handler (static)
 */
// Exit
void AG_GUI_BASE::OnSelectExit(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	gui->Exit();
}
////////////////////////////////////////
// Reset
void AG_GUI_BASE::OnSelectReset(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	gui->Reset();
}
// SpecialReset
void AG_GUI_BASE::OnSelectSpecialReset(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	gui->SpecialReset();
}
// WarmReset
void AG_GUI_BASE::OnSelectWarmReset(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	gui->WarmReset();
}

// change dipswitch
void AG_GUI_BASE::OnSelectDipswitch(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int bit = AG_INT(2);
	gui->Dipswitch(bit);
}
// update dipswitch
void AG_GUI_BASE::OnUpdateDipswitch(AG_Event *event)
{
	AG_MenuItem *mi = (AG_MenuItem *)AG_SENDER();
//	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int bit = AG_INT(2);

	AG_GUI_MENU_CHECK(mi, config.dipswitch & (1 << bit));
}

// change pause
void AG_GUI_BASE::OnSelectPause(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	gui->Pause();
}
// update pause
void AG_GUI_BASE::OnUpdatePause(AG_Event *event)
{
	AG_MenuItem *mi = (AG_MenuItem *)AG_SENDER();
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);

	AG_GUI_MENU_CHECK(mi, gui->emu->get_pause(2));
}

// change cpu power
void AG_GUI_BASE::OnSelectCPUPower(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int num = AG_INT(2);
	gui->CPUPower(num);
}
// update cpu power
void AG_GUI_BASE::OnUpdateCPUPower(AG_Event *event)
{
	AG_MenuItem *mi = (AG_MenuItem *)AG_SENDER();
//	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int num = AG_INT(2);

	AG_GUI_MENU_CHECK(mi, config.cpu_power == num);
}
// show Paste Text dialog (auto key)
void AG_GUI_BASE::OnSelectStartAutoKey(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	char *ext = AG_STRING(2);

	gui->filebox->CreateLoad(AG_FILE_DLG::AUTOKEY, "Open Auto key file"
		, ext, config.initial_autokey_path, NULL);
}
// update paste text (auto key)
void AG_GUI_BASE::OnUpdateStartAutoKey(AG_Event *event)
{
	AG_MenuItem *mi = (AG_MenuItem *)AG_SENDER();
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);

	if (gui->emu->now_auto_key()) {
		mi->state = 0;	// disable
	} else {
		mi->state = 1;	// enable
	}
}
// stop auto key
void AG_GUI_BASE::OnSelectStopAutoKey(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);

	gui->emu->stop_auto_key();
}

////////////////////////////////////////
// Show Data Recorder Load Dialog
void AG_GUI_BASE::OnSelectLoadDataRec(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
//	char *ext = AG_STRING(2);

	gui->ShowLoadDataRecDialog();
}
// update data rec load
void AG_GUI_BASE::OnUpdateLoadDataRec(AG_Event *event)
{
	AG_MenuItem *mi = (AG_MenuItem *)AG_SENDER();
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);

	AG_GUI_MENU_CHECK(mi, gui->emu->datarec_opened(true));
}
// Show Data Recorder Save Dialog
void AG_GUI_BASE::OnSelectSaveDataRec(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
//	char *ext = AG_STRING(2);

	gui->ShowSaveDataRecDialog();
}
// update data rec save
void AG_GUI_BASE::OnUpdateSaveDataRec(AG_Event *event)
{
	AG_MenuItem *mi = (AG_MenuItem *)AG_SENDER();
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);

	AG_GUI_MENU_CHECK(mi, gui->emu->datarec_opened(false));
}
// Close Data Recorder file
void AG_GUI_BASE::OnSelectCloseDataRec(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);

	gui->emu->close_datarec();
}
// select Recent file (Datarec)
void AG_GUI_BASE::OnSelectRecentDataRec(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	char file_path[_MAX_PATH];

	memset(file_path, '\0', sizeof(file_path));
	strncpy(file_path, AG_STRING(2), _MAX_PATH);

	gui->LoadDataRec(file_path);
}

////////////////////////////////////////
// Show Floppy Disk Open Dialog
void AG_GUI_BASE::OnSelectOpenFloppyDisk(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int   drv = AG_INT(2);
//	char *ext = AG_STRING(3);

	gui->ShowOpenFloppyDiskDialog(drv);
}
// update floppy open
void AG_GUI_BASE::OnUpdateOpenFloppyDisk(AG_Event *event)
{
	AG_MenuItem *mi = (AG_MenuItem *)AG_SENDER();
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int   drv = AG_INT(2);

	AG_GUI_MENU_CHECK(mi, gui->emu->disk_inserted(drv));
}
// Close Floppy Disk file
void AG_GUI_BASE::OnSelectCloseFloppyDisk(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int   drv = AG_INT(2);

//	gui->emu->close_disk(drv);
	gui->CloseFloppyDisk(drv);
}
// select Recent file (floppy disk)
void AG_GUI_BASE::OnSelectRecentFloppyDisk(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int   drv = AG_INT(2);
	char file_path[_MAX_PATH];

	memset(file_path, '\0', sizeof(file_path));
	strncpy(file_path, AG_STRING(3), _MAX_PATH);

	gui->OpenFloppyDisk(drv, file_path, 0, 0, true);
}
void AG_GUI_BASE::OnUpdateMultiVolumeList(AG_Event *event)
{
	AG_MenuItem *mi = (AG_MenuItem *)AG_SENDER();
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int   drv = AG_INT(2);

	gui->update_multi_volume_list(mi, drv);
}
/// select bank number of multi volume (floppy disk)
void AG_GUI_BASE::OnSelectVolumeFloppyDisk(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int   drv = AG_INT(2);
	int   bank_num = AG_INT(3);

	gui->OpenFloppyDiskSelectedVolume(drv, bank_num);
}
/// update bank number of multi volume
void AG_GUI_BASE::OnUpdateVolumeFloppyDisk(AG_Event *event)
{
//	AG_MenuItem *mi = (AG_MenuItem *)AG_SENDER();
//	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
//	int   drv = AG_INT(2);
//	int   bank_num = AG_INT(3);
//
//	AG_GUI_MENU_CHECK(mi, gui->InsertedFloppyDisk(drv));
}

////////////////////////////////////////
// Change Frame Rate
void AG_GUI_BASE::OnSelectFrameRate(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int num = AG_INT(2);

	config.fps_no = num;
	gui->emu->update_config();
}
// update Frame Rate
void AG_GUI_BASE::OnUpdateFrameRate(AG_Event *event)
{
	AG_MenuItem *mi = (AG_MenuItem *)AG_SENDER();
//	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int num = AG_INT(2);

	AG_GUI_MENU_CHECK(mi, config.fps_no == num);
}
// Change Screen Record Size
void AG_GUI_BASE::OnSelectScreenRecordSize(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int num = AG_INT(2);

	gui->emu->resize_rec_video(num);
}
// update Screen Record Size
void AG_GUI_BASE::OnUpdateScreenRecordSize(AG_Event *event)
{
	AG_MenuItem *mi = (AG_MenuItem *)AG_SENDER();
//	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int num = AG_INT(2);

	AG_GUI_MENU_CHECK(mi, config.screen_video_size == num);
}
// Change Screen Capture
void AG_GUI_BASE::OnSelectScreenCapture(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);

	gui->emu->capture_screen();
}
// Change Window Mode
void AG_GUI_BASE::OnSelectWindowMode(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int num = AG_INT(2);

	gui->ChangeWindowMode(num);
}

// update Window Mode
void AG_GUI_BASE::OnUpdateWindowMode(AG_Event *event)
{
	AG_MenuItem *mi = (AG_MenuItem *)AG_SENDER();
	int num = AG_INT(2);

	AG_GUI_MENU_CHECK(mi, config.window_mode == num);
}
// Change Screen Mode
void AG_GUI_BASE::OnSelectScreenMode(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int num = AG_INT(2);

	gui->ChangeScreenMode(num);
}
// update Screen Mode
void AG_GUI_BASE::OnUpdateScreenMode(AG_Event *event)
{
	AG_MenuItem *mi = (AG_MenuItem *)AG_SENDER();
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int num = AG_INT(2);

	AG_GUI_MENU_CHECK(mi, config.window_mode == num + 8);
	AG_GUI_MENU_ENABLE(mi, gui->emu->is_fullscreen() == 0);
}
// Change Scan Line
void AG_GUI_BASE::OnSelectScanLine(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);

	config.scan_line = !config.scan_line;
	gui->emu->update_config();
}
// update Scan Line
void AG_GUI_BASE::OnUpdateScanLine(AG_Event *event)
{
	AG_MenuItem *mi = (AG_MenuItem *)AG_SENDER();
//	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);

	AG_GUI_MENU_CHECK(mi, config.scan_line);
}

////////////////////////////////////////
// create Volume dialog
void AG_GUI_BASE::OnSelectSoundVolume(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);

	gui->CreateVolumeBox();
}
// Change Sound Start Record
void AG_GUI_BASE::OnSelectSoundStartRecord(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);

	gui->emu->start_rec_sound();
}
// update Sound Start Record
void AG_GUI_BASE::OnUpdateSoundStartRecord(AG_Event *event)
{
	AG_MenuItem *mi = (AG_MenuItem *)AG_SENDER();
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);

	AG_GUI_MENU_ENABLE(mi, !gui->emu->now_rec_sound());
}
// Change Sound Start Record
void AG_GUI_BASE::OnSelectSoundStopRecord(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);

	gui->emu->stop_rec_sound();
}
// update Sound Start Record
void AG_GUI_BASE::OnUpdateSoundStopRecord(AG_Event *event)
{
	AG_MenuItem *mi = (AG_MenuItem *)AG_SENDER();
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);

	AG_GUI_MENU_ENABLE(mi, gui->emu->now_rec_sound());
}
// Change Sound Rate
void AG_GUI_BASE::OnSelectSoundRate(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int num = AG_INT(2);

	config.sound_frequency = num;
	gui->emu->update_config();
}
// update Sound Rate
void AG_GUI_BASE::OnUpdateSoundRate(AG_Event *event)
{
	AG_MenuItem *mi = (AG_MenuItem *)AG_SENDER();
//	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int num = AG_INT(2);

	AG_GUI_MENU_CHECK(mi, config.sound_frequency == num);
}
// Change Sound Latency
void AG_GUI_BASE::OnSelectSoundLatency(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int num = AG_INT(2);

	config.sound_latency = num;
	gui->emu->update_config();
}
// update Sound Latency
void AG_GUI_BASE::OnUpdateSoundLatency(AG_Event *event)
{
	AG_MenuItem *mi = (AG_MenuItem *)AG_SENDER();
//	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int num = AG_INT(2);

	AG_GUI_MENU_CHECK(mi, config.sound_latency == num);
}

////////////////////////////////////////
// Show Print Data Save Dialog
void AG_GUI_BASE::OnSelectSavePrinter(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int drv = AG_INT(2);
	char *ext = AG_STRING(3);

	gui->filebox->CreateSave(AG_FILE_DLG::PRINTER, "Print Data"
		, ext, config.initial_printer_path, NULL, (void *)drv);
}
// update Save Printer
void AG_GUI_BASE::OnUpdateSavePrinter(AG_Event *event)
{
	AG_MenuItem *mi = (AG_MenuItem *)AG_SENDER();
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int drv = AG_INT(2);

	int size = gui->emu->get_printer_buffer_size(drv);
	AG_GUI_MENU_ENABLE(mi, size > 0);
}
// Clear Print Data
void AG_GUI_BASE::OnSelectClearPrinter(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int drv = AG_INT(2);

	gui->emu->clear_printer(drv);
}
// update Clear Printer
void AG_GUI_BASE::OnUpdateClearPrinter(AG_Event *event)
{
	AG_MenuItem *mi = (AG_MenuItem *)AG_SENDER();
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int drv = AG_INT(2);

	int size = gui->emu->get_printer_buffer_size(drv);
	AG_GUI_MENU_ENABLE(mi, size > 0);
}

///
/// process user event
void AG_GUI_BASE::OnProcessUserEvent(AG_Event *event)
{
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int id = AG_INT(2);

	gui->ProcessUserEvent(id);
}

////////////////////////////////////////
// update Recent file
void AG_GUI_BASE::OnUpdateRecent(AG_Event *event)
{
	AG_MenuItem *mi = (AG_MenuItem *)AG_SENDER();
	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);
	int type = AG_INT(2);
	int drv = AG_INT(3);

	gui->update_recent_list(mi, type, drv);
}

////////////////////////////////////////
// select About dialog
void AG_GUI_BASE::OnSelectAbout(AG_Event *event)
{
//	AG_GUI_BASE *gui = (AG_GUI_BASE *)AG_PTR(1);

	const SDL_version *sdl_ver;
	AG_AgarVersion agar_ver;

	sdl_ver = SDL_Linked_Version();
	AG_GetVersion(&agar_ver);

	AG_TextMsg(AG_MSG_INFO,
		"%s Emulator\n"
		"  Version: %s \"%s\"\n\n"
		"  SDL Version: %d.%d.%d\n"
		"  Agar Version: %d.%d.%d (%s)"
		,DEVICE_NAME
		,VM_VERSION
		,PLATFORM
		,sdl_ver->major, sdl_ver->minor, sdl_ver->patch
		,agar_ver.major, agar_ver.minor, agar_ver.patch, agar_ver.release
	);

}

