/** @file sdl_input.cpp

	Skelton for retropc emulator
	SDL edition

	@author Sasaji
	@date   2012.02.21

	@brief [ sdl input ]

	@note
	This code is based on win32_input.cpp of the Common Source Code Project.
	Author : Takeda.Toshiya
*/

#include "emu.h"
#include "vm/vm.h"
#include "fifo.h"
#include "fileio.h"
#include "sdl_input.h"

#define KEY_KEEP_FRAMES 3

void EMU::EMU_INPUT()
{
	vm_key_status = NULL;
	vm_key_status_size = 0;

	pressed_global_key = false;
}

void EMU::initialize_input()
{
	out_debug(_T("EMU::initialize_input"));

	// initialize status
	memset(key_status, 0, sizeof(key_status));
	memset(joy_status, 0, sizeof(joy_status));
	memset(mouse_status, 0, sizeof(mouse_status));

	for(int i=0; i<2; i++) {
		joy[i] = NULL;
		joy_enabled[i] = false;
	}
	use_joystick = (config.show_dlg & 4) ? true : false;
	// initialize joysticks
	if (SDL_InitSubSystem(SDL_INIT_JOYSTICK)) {
		out_log(LOG_WARN, _T("SDL_InitSubSystem(SDL_INIT_JOYSTICK): %s."), SDL_GetError());
	}
	initialize_joystick();

	// mouse emulation is disenabled
	mouse_enabled = false;

#ifdef USE_AUTO_KEY
	// initialize autokey
	autokey_buffer = new FIFO(65536);
	autokey_buffer->clear();
	autokey_phase = 31;	// wait a few sec.
	autokey_shift = 0;
	autokey_enabled = false;
#endif
	lost_focus = false;
	key_mod = 0;
}

void EMU::initialize_joystick()
{
#ifdef USE_JOYSTICK
	for(int i = 0; i < 2; i++) {
		joy_xmin[i] = -16384;
		joy_xmax[i] = 16384;
		joy_ymin[i] = -16384;
		joy_ymax[i] = 16384;
	}
	SDL_JoystickEventState(SDL_IGNORE);
	reset_joystick();
#endif
}

void EMU::reset_joystick()
{
#ifdef USE_JOYSTICK
	if (!use_joystick) return;

	joy_num = SDL_NumJoysticks();
	for(int i = 0; i < joy_num && i < 2; i++) {
		if (!SDL_JoystickOpened(i)) {
			if ((joy[i] = SDL_JoystickOpen(i)) != NULL) {
				joy_enabled[i] = true;
			}
		}
	}
#endif
}

void EMU::release_input()
{
	// release mouse
	if(mouse_enabled) {
		disenable_mouse();
	}
	// release joystick
	release_joystick();

#ifdef USE_AUTO_KEY
	// release autokey buffer
	if(autokey_buffer) {
		autokey_buffer->release();
		delete autokey_buffer;
	}
#endif
}

void EMU::release_joystick()
{
#ifdef USE_JOYSTICK
	// release joystick
	for(int i = 0; i < 2; i++) {
		if (joy[i] != NULL) {
			SDL_JoystickClose(joy[i]);
		}
		joy[i] = NULL;
		joy_enabled[i] = false;
	}
#endif
}

void EMU::update_input()
{
//	emu->out_debug("EMU::update_input");

	// release keys
#ifdef USE_AUTO_KEY
	if(lost_focus && !autokey_enabled) {
#else
	if(lost_focus) {
#endif
		// we lost key focus so release all pressed keys
		for(int i = 0; i < KEY_STATUS_SIZE; i++) {
			if(key_status[i] & 0x80) {
				key_status[i] &= 0x7f;
#ifdef NOTIFY_KEY_DOWN
				if(!key_status[i]) {
					vm->key_up(i);
				}
#endif
			}
		}
	}
	else {
		for(int i = 0; i < KEY_STATUS_SIZE; i++) {
			if(key_status[i] & 0x7f) {
				key_status[i] = (key_status[i] & 0x80) | ((key_status[i] & 0x7f) - 1);
#ifdef NOTIFY_KEY_DOWN
				if(!key_status[i]) {
					vm->key_up(i);
				}
#endif
			}
		}
	}
	lost_focus = false;

#ifdef USE_JOYSTICK
	update_joystick();
#endif
#ifdef USE_KEY_TO_JOY
	// emulate joystick #1 with keyboard
	if(key_status[0x26]) joy_status[0] |= 0x01;	// up
	if(key_status[0x28]) joy_status[0] |= 0x02;	// down
	if(key_status[0x25]) joy_status[0] |= 0x04;	// left
	if(key_status[0x27]) joy_status[0] |= 0x08;	// right
#endif

	// update mouse status
	memset(mouse_status, 0, sizeof(mouse_status));
	if(mouse_enabled) {
		// get current status
		Point pt;
		Uint8 mstat = SDL_GetMouseState(&(pt.x),&(pt.y));

		mouse_status[0]  = pt.x - display_width / 2;
		mouse_status[1]  = pt.y - display_height / 2;
		mouse_status[2] =  (mstat & SDL_BUTTON(SDL_BUTTON_LEFT) ? 1 : 0);
		mouse_status[2] |= (mstat & SDL_BUTTON(SDL_BUTTON_RIGHT) ? 2 : 0);
		mouse_status[2] |= (mstat & SDL_BUTTON(SDL_BUTTON_MIDDLE) ? 4 : 0);
		// move mouse cursor to the center of window
		if(!(mouse_status[0] == 0 && mouse_status[1] == 0)) {
			pt.x = display_width / 2;
			pt.y = display_height / 2;
			SDL_WarpMouse(pt.x, pt.y);
		}
	}
#ifdef USE_LIGHTPEN
	if(true) {
		// get current status
		Point pt;
		Uint8 mstat = SDL_GetMouseState(&(pt.x),&(pt.y));
		pt.x = screen_width * (pt.x - screen_dest_x) / stretched_width;
		pt.y = screen_height * (pt.y - screen_dest_y) / stretched_height;

		mouse_status[0] = pt.x;
		mouse_status[1] = pt.y;
		mouse_status[2] = (mstat & SDL_BUTTON(SDL_BUTTON_LEFT) ? 1 : 0) | (mstat & SDL_BUTTON(SDL_BUTTON_RIGHT) ? 2 : 0);
	}
#endif
}

void EMU::update_autokey()
{
#ifdef USE_AUTO_KEY
	// auto key
	switch(autokey_phase) {
	case 1:
		if(autokey_buffer && !autokey_buffer->empty()) {
#ifdef _BML3MK5
			// update graph key status
			int graph = autokey_buffer->read_not_remove(0) & AUTO_KEY_GRAPH_MASK;
			if(graph && !autokey_shift) {
#ifdef USE_AUTO_KEY_VM_CODE
				vm_key_down(AUTO_KEY_GRAPH);
#else
				key_down(AUTO_KEY_GRAPH, 0);
#endif
				out_debug(_T("autokey phase:% 2d: graph down"), autokey_phase);
				autokey_shift = graph;
				autokey_phase++;
				break;
			}
#endif
			// update shift key status
			int shift = autokey_buffer->read_not_remove(0) & AUTO_KEY_SHIFT_MASK;
			if(shift && !autokey_shift) {
#ifdef USE_AUTO_KEY_VM_CODE
				vm_key_down(AUTO_KEY_SHIFT);
#else
				key_down(AUTO_KEY_SHIFT, KMOD_SHIFT);
#endif
				out_debug(_T("autokey phase:% 2d: shift down"), autokey_phase);
				autokey_shift = shift;
				autokey_phase++;
				break;
			}
			autokey_shift = 0;
		}
		autokey_phase++;
		break;
	case 2:
		if(autokey_buffer && !autokey_buffer->empty()) {
			autokey_code = autokey_buffer->read_not_remove(0) & AUTO_KEY_MASK;
#ifdef USE_AUTO_KEY_VM_CODE
			vm_key_down(autokey_code);
#else
			autokey_code = key_down(autokey_code, 0);
#endif
			out_debug(_T("autokey phase:% 2d: key down %x"), autokey_phase, autokey_code);
		}
		autokey_phase++;
		break;
	case USE_AUTO_KEY:
		if(!autokey_buffer || autokey_buffer->empty()) {
			autokey_phase = 30;
		} else {
			// wait response from vm
			out_debug(_T("autokey phase:% 2d: wait "), autokey_phase);
		}
		break;
	case USE_AUTO_KEY + 2:
		if(autokey_shift & AUTO_KEY_SHIFT_MASK) {
#ifdef USE_AUTO_KEY_VM_CODE
			vm_key_up(AUTO_KEY_SHIFT);
#else
			key_up(AUTO_KEY_SHIFT, 0);
#endif
			out_debug(_T("autokey phase:% 2d: shift up:"), autokey_phase);
		}
#ifdef _BML3MK5
		if(autokey_shift & AUTO_KEY_GRAPH_MASK) {
#ifdef USE_AUTO_KEY_VM_CODE
			vm_key_up(AUTO_KEY_GRAPH);
#else
			key_up(AUTO_KEY_GRAPH, 0);
#endif
			out_debug(_T("autokey phase:% 2d: graph up"), autokey_phase);
		}
#endif
		autokey_shift = 0;
		if(autokey_buffer && !autokey_buffer->empty()) {
#ifdef USE_AUTO_KEY_VM_CODE
			vm_key_up(autokey_buffer->read_not_remove(0) & AUTO_KEY_MASK);
#else
			key_up(autokey_buffer->read_not_remove(0) & AUTO_KEY_MASK, 0);
#endif
		}
		out_debug(_T("autokey phase:% 2d: key up %x"), autokey_phase, autokey_code);
		autokey_code = 0;
		autokey_phase++;
		break;
	case USE_AUTO_KEY + 3:
		if(!autokey_buffer || autokey_buffer->empty()) {
			autokey_phase = 30;
		} else {
			// wait response from vm
			out_debug(_T("autokey phase:% 2d: wait "), autokey_phase);
		}
		break;
	case USE_AUTO_KEY + 4:
		if(autokey_buffer && !autokey_buffer->empty()) {
			// wait enough while vm analyzes one line
			int code = autokey_buffer->read();
			out_debug(_T("autokey phase:% 2d: code %x"), autokey_phase, code);
			if(code == AUTO_KEY_RETURN) {
				autokey_phase++;	// wait 30 frames
				break;
			}
		}
		// through
	case 30:
	case 150:
		if(autokey_buffer && !autokey_buffer->empty()) {
			autokey_phase = 1;
		}
		else {
			stop_auto_key();
			autokey_phase = 0;
		}
		break;
	default:
		if(autokey_phase) {
			autokey_phase++;
		}
		break;
	}
#endif
}

void EMU::recognized_key(uint16 key_code)
{
#ifdef USE_AUTO_KEY
	if(autokey_phase == 0) return;
	if(autokey_phase == USE_AUTO_KEY && autokey_code) {
		out_debug(_T("recognized_key: phase:%d auto_key:%x vm_key:%x"),autokey_phase,autokey_code,key_code);
		if (autokey_code == key_code) {
			autokey_phase = USE_AUTO_KEY + 1;
		}
	}
	if(autokey_phase == USE_AUTO_KEY + 3 && key_code == AUTO_KEY_NONE) {
		out_debug(_T("recognized_key: phase:%d auto_key:%x vm_key:%x"),autokey_phase,autokey_code,key_code);
		autokey_phase =  USE_AUTO_KEY + 4;
	}
#endif
}

void EMU::update_joystick()
{
#ifdef USE_JOYSTICK
	memset(joy_status, 0, sizeof(joy_status));

	if (!use_joystick) return;

	// update joystick status
	SDL_JoystickUpdate();
	for(int i = 0; i < 2; i++) {
		if(joy_enabled[i] == true) {
			if (SDL_JoystickOpened(i)) {
				int x = SDL_JoystickGetAxis(joy[i], 0);
				int y = SDL_JoystickGetAxis(joy[i], 1);
				if(y < joy_ymin[i]) joy_status[i] |= 0x01;
				if(y > joy_ymax[i]) joy_status[i] |= 0x02;
				if(x < joy_xmin[i]) joy_status[i] |= 0x04;
				if(x > joy_xmax[i]) joy_status[i] |= 0x08;
				int nbuttons = SDL_JoystickNumButtons(joy[i]);
				for(int n=0; n<nbuttons && n<28; n++) {
					if (SDL_JoystickGetButton(joy[i], n)) {
						joy_status[i] |= (0x10 << n);
					}
				}
			} else {
				joy_enabled[i] = false;
			}
		}
	}
#endif
}

int EMU::key_down(int code, uint16 mod)
{
	if (!translate_key_down(code, mod)) {
		key_status[code] = 0x80;
	}
#ifdef NOTIFY_KEY_DOWN
	vm->key_down(code);
#endif
	return code;
}

void EMU::key_up(int code, uint16 mod)
{
	if(key_status[code]) {
		key_status[code] &= 0x7f;
#ifdef NOTIFY_KEY_DOWN
		if(!key_status[code]) {
			vm->key_up(code);
		}
#endif
	}
}

int EMU::vm_key_down(int code)
{
#ifdef _BML3MK5
	code -= 0x80;
#endif
	if (vm_key_status && 0 <= code && code < vm_key_status_size) {
		vm_key_status[code] = 1;
	}
	return code;
}
void EMU::vm_key_up(int code)
{
#ifdef _BML3MK5
	code -= 0x80;
#endif
	if (vm_key_status && 0 <= code && code < vm_key_status_size) {
		vm_key_status[code] = 0;
	}
}

#ifdef USE_BUTTON
void EMU::press_button(int num)
{
	int code = buttons[num].code;

	if(code) {
		key_down(code, false);
		key_status[code] = KEY_KEEP_FRAMES * FRAME_SPLIT_NUM;
	}
	else {
		// code=0: reset virtual machine
		vm->reset();
	}
}
#endif

void EMU::enable_mouse()
{
	// enable mouse emulation
	if(!mouse_enabled) {
		// hide mouse cursor
		SDL_ShowCursor(SDL_DISABLE);
		// move mouse cursor to the center of window
		Point pt;
		pt.x = display_width / 2;
		pt.y = display_height / 2;
		SDL_WarpMouse(pt.x, pt.y);
	}
	mouse_enabled = true;
}

void EMU::disenable_mouse()
{
	// disenable mouse emulation
	if(mouse_enabled) {
		SDL_ShowCursor(SDL_ENABLE);
	}
	mouse_enabled = false;
}

void EMU::toggle_mouse()
{
	// toggle mouse enable / disenable
	if(mouse_enabled) {
		disenable_mouse();
	}
	else {
		enable_mouse();
	}
}

#ifdef USE_AUTO_KEY
bool EMU::start_auto_key()
{
	bool rc = true;
#if 0
	stop_auto_key();

	autokey_code = 0;
	if(OpenClipboard(NULL)) {
		HANDLE hClip = GetClipboardData(CF_TEXT);
		if(hClip) {
			autokey_buffer->clear();
			int prev_code = 0;
			char* buf = (char*)GlobalLock(hClip);
			int size = (int)strlen(buf), prev_kana = 0;
			parse_auto_key(buf, size, prev_code, prev_kana);
			if(prev_kana) {
				autokey_buffer->write(AUTO_KEY_KANA);
#ifdef _BML3MK5
				if (prev_kana == 0x1000) autokey_buffer->write(AUTO_KEY_KANA);
#endif
			}
			GlobalUnlock(hClip);

			autokey_phase = (autokey_phase > 30 ? autokey_phase : 1);
			autokey_shift = 0;
			autokey_enabled = true;
		}
		CloseClipboard();
	} else {
		rc = false;
	}

	update_config();
#endif
	return rc;
}

bool EMU::start_auto_key(_TCHAR *file_path)
{
	FILEIO *fp;
	char buf[1025];
	bool rc = true;

	stop_auto_key();

	autokey_code = 0;
	fp = new FILEIO();
	int prev_kana = 0;
	if (fp->Fopen(file_path, FILEIO_READ_BINARY)) {
		autokey_buffer->clear();
		int prev_code = 0;
		int size = 0;
		while((size = (int)fp->Fread(buf, sizeof(char), 1024)) > 0) {
			parse_auto_key(buf, size, prev_code, prev_kana);
		}
		fp->Fclose();

		if(prev_kana) {
			autokey_buffer->write(AUTO_KEY_KANA);
#ifdef _BML3MK5
			if (prev_kana == 0x1000) autokey_buffer->write(AUTO_KEY_KANA);
#endif
		}

		autokey_phase = (autokey_phase > 30 ? autokey_phase : 1);
		autokey_shift = 0;
		autokey_enabled = true;
		_tcsncpy(config.opened_autokey_path, file_path, _MAX_PATH);
	} else {
		out_log(LOG_ERROR, _T("Auto key file couldn't be opened."));
		rc = false;
	}
	delete fp;

	copy_initial_path(config.initial_autokey_path, file_path);

	update_config();

	return rc;
}

void EMU::parse_auto_key(char *buf, int size, int &prev_code, int &prev_kana)
{
	for(int i = 0; i < size; i++) {
		int code = buf[i] & 0xff;
#ifdef _BML3MK5
		if(prev_code == 0xd && code == 0xa) {
			prev_code = code;
			continue;	// cr-lf
		}
#else
		if((0x81 <= code && code <= 0x9f) || 0xe0 <= code) {
			i++;	// kanji ?
			continue;
		}
		else if(code == 0xa) {
			continue;	// cr-lf
		}
#endif
		if((code = autokey_table[code]) != 0) {
#ifdef _BML3MK5
			int kana = code & 0x3000;
			if(kana - prev_kana == 0x1000) {
				autokey_buffer->write(AUTO_KEY_KANA);
			} else if(kana - prev_kana == 0x2000) {
				autokey_buffer->write(AUTO_KEY_KANA);
				autokey_buffer->write(AUTO_KEY_KANA);
			}
			if(prev_kana - kana == 0x2000) {
				autokey_buffer->write(AUTO_KEY_KANA);
			} else if(prev_kana - kana == 0x1000) {
				autokey_buffer->write(AUTO_KEY_KANA);
				autokey_buffer->write(AUTO_KEY_KANA);
			}
#else
			int kana = code & 0x200;
			if(prev_kana != kana) {
				autokey_buffer->write(AUTO_KEY_KANA);
			}
#endif
			prev_kana = kana;
#if defined(USE_AUTO_KEY_NO_CAPS)
			if((code & AUTO_KEY_SHIFT_MASK) && !(code & (AUTO_KEY_UPPER_MASK | AUTO_KEY_LOWER_MASK))) {
#elif defined(USE_AUTO_KEY_CAPS)
			if(code & (AUTO_KEY_SHIFT_MASK | AUTO_KEY_LOWER_MASK)) {
#else
			if(code & (AUTO_KEY_SHIFT_MASK | AUTO_KEY_UPPER_MASK)) {
#endif
				autokey_buffer->write((code & AUTO_KEY_MASK) | AUTO_KEY_SHIFT_MASK);
			}
#ifdef _BML3MK5
			else if (code & AUTO_KEY_GRAPH_MASK) {
				autokey_buffer->write((code & AUTO_KEY_MASK) | AUTO_KEY_GRAPH_MASK);
			}
#endif
			else {
				autokey_buffer->write(code & AUTO_KEY_MASK);
			}
		}
		prev_code =code;
	}
}

void EMU::stop_auto_key()
{
#ifndef _BML3MK5
	if(autokey_shift) {
		key_up(AUTO_KEY_SHIFT);
	}
#else
	if(autokey_shift) {
		vm_key_up(AUTO_KEY_SHIFT);
	}
	if(autokey_shift) {
		vm_key_up(AUTO_KEY_GRAPH);
	}
	if (vm_key_status) {
		memset(vm_key_status, 0, sizeof(uint8) * vm_key_status_size);
	}
#endif
	autokey_phase = (autokey_phase > 30 ? autokey_phase : 0);
	autokey_shift = 0;
	autokey_enabled = false;

	memset(config.opened_autokey_path, 0, sizeof(config.opened_autokey_path));

	update_config();
}
#endif
