/** @file sdl_main.cpp

	Skelton for retropc emulator
	SDL edition

	@author Sasaji
	@date   2012.2.15

	@brief [ sdl_main ]

	@note
	This code is based on winmain.cpp of the Common Source Code Project.
	Author : Takeda.Toshiya
*/

#include <sys/types.h>
#include <agar/core.h>
#include "common.h"
#include "config.h"
#include "emu.h"
#include "gui/ag_gui.h"
#include "gui/ag_file_dlg.h"

#define MIN_SKIP_FRAMES 0
#define MAX_SKIP_FRAMES 10

#define MIN_INTERVAL	4

#define CALC_NEXT_INTERVAL(now_ms, interval_ms, delta_ms) { \
	interval_ms = now_ms / FRAMES_PER_SEC; \
	now_ms = now_ms - (interval_ms * FRAMES_PER_SEC) + delta_ms; \
}

// variable
EMU *emu = NULL;
AG_GUI_BASE *gui = NULL;


int fskip_remain = 0;
static const int fps[6] = {60, 30, 20, 15, 12, 10};
static const int fskip[6] = {1, 2, 3, 4, 5, 6};
int rec_fps_no = -1;


bool working = true;
bool gui_process = false;

int *key_mod = NULL;

static char app_path[_MAX_PATH];
static char app_name[_MAX_PATH];
static char ini_path[_MAX_PATH];
static char ini_file[_MAX_PATH];
static char res_path[_MAX_PATH];
static char *tape_file;
static char *disk_file;
static char *state_file;
static char *autokey_file;
static char *reckey_file;

int desktop_width = 0;
int desktop_height = 0;

// prototype
static void exit_program(void);

static void main_event_loop();
static void vm_event_loop();

static void vm_event_proc(SDL_Event *);

static int get_options(int, char *[]);
static void open_recent_file();
int check_supported_file(char *file_path);

/**
 *	main
 */
int main(int argc, char* argv[])
{
#if defined(_DEBUG) && defined(_WIN32) && !defined(__MINGW32__)
	_CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif

	char buf[_MAX_PATH];

	const SDL_VideoInfo *video = NULL;
//	char icon_file[_MAX_PATH];
#ifdef _WIN32
	CoInitialize(NULL);
#endif

	// get and parse options
	get_options(argc, argv);

	// create a instance of emulation core
	emu = new EMU(app_path, ini_path, res_path);

	// load config
	load_config(ini_file);

	gui = new AG_GUI(emu);
	if (gui->InitCore() == -1) {
		delete gui;
		exit(1);
	}
	emu->set_gui(gui);

	// disable keylock
//	SDL_putenv("SDL_DISABLE_LOCK_KEYS=1");

	// init SDL
	Uint32 init_flags = SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_TIMER;
#ifndef _DEBUG
	init_flags |= SDL_INIT_NOPARACHUTE;
#endif
	if(SDL_Init(init_flags)) {
		emu->out_log(LOG_ERROR, _T("SDL_Init: %s."), SDL_GetError());
		exit(1);
	}
	atexit(exit_program);

	// UNICODE support
	if (SDL_EnableUNICODE(1) == -1) {
		emu->out_log(LOG_WARN, _T("SDL_EnableUNICODE: %s."), SDL_GetError());
	}

#if !(defined(__APPLE__) && defined(__MACH__))
	// window icon
	sprintf(buf, "%s/res/%s.bmp", app_path, CONFIG_NAME);
	SDL_WM_SetIcon(SDL_LoadBMP(buf), NULL);
#endif

	// get current video mode
	if((video = SDL_GetVideoInfo()) == NULL) {
		emu->out_log(LOG_ERROR, _T("SDL_GetVideoInfo: %s."), SDL_GetError());
		exit(1);
	}
#ifdef _DEBUG_LOG
	emu->out_debug(_T("video->hw_available:%d"),video->hw_available);
	emu->out_debug(_T("video->wm_available:%d"),video->wm_available);
	emu->out_debug(_T("video->blit_hw:%d"),video->blit_hw);
	emu->out_debug(_T("video->blit_hw_CC:%d"),video->blit_hw_CC);
	emu->out_debug(_T("video->blit_hw_A:%d"),video->blit_hw_A);
	emu->out_debug(_T("video->blit_sw:%d"),video->blit_sw);
	emu->out_debug(_T("video->blit_sw_CC:%d"),video->blit_sw_CC);
	emu->out_debug(_T("video->blit_sw_A:%d"),video->blit_sw_A);
	emu->out_debug(_T("video->blit_fill:%d"),video->blit_fill);
	emu->out_debug(_T("video->video_mem:%d"),video->video_mem);
	emu->out_debug(_T("video->vfmt->BitsPerPixel:%d"),video->vfmt->BitsPerPixel);
	emu->out_debug(_T("video->vfmt->BytesPerPixel:%d"),video->vfmt->BytesPerPixel);
	emu->out_debug(_T("video->vfmt->Rmask:%x, Gmask:%x, Bmask:%x, Amask:%x"),video->vfmt->Rmask,video->vfmt->Gmask,video->vfmt->Bmask,video->vfmt->Amask);
	emu->out_debug(_T("video->vfmt->Rshift:%d, Gshift:%d, Bshift:%d, Ashift:%d"),video->vfmt->Rshift,video->vfmt->Gshift,video->vfmt->Bshift,video->vfmt->Ashift);
	emu->out_debug(_T("video->vfmt->Rloss:%d, Gloss:%d, Bloss:%d, Aloss:%d"),video->vfmt->Rloss,video->vfmt->Gloss,video->vfmt->Bloss,video->vfmt->Aloss);
	emu->out_debug(_T("video->vfmt->colorkey:%d"),video->vfmt->colorkey);
	emu->out_debug(_T("video->vfmt->alpha:%d"),video->vfmt->alpha);
	emu->out_debug(_T("video->current_w:%d"),video->current_w);
	emu->out_debug(_T("video->current_h:%d"),video->current_h);
#endif
	desktop_width = video->current_w;
	desktop_height = video->current_h;

	Uint32 video_flags = 0;
	if (video->hw_available) {
		video_flags |= SDL_HWSURFACE;
	}
#if (defined(__APPLE__) && defined(__MACH__))
	// for mac
	video_flags |= SDL_ASYNCBLIT;
#endif
#ifdef USE_OPENGL
	emu->set_use_opengl(config.use_opengl);
#endif
	// create window
	emu->set_desktop_size(desktop_width, desktop_height);
	emu->enum_window_mode(desktop_width, desktop_height);
	emu->enum_screen_mode(video->vfmt, video_flags);
	emu->find_screen_mode();
	emu->set_window(config.window_mode, desktop_width, desktop_height);

	if (!emu->create_screen(config.screen_width, config.screen_height, video_flags)) {
		exit(1);
	}
	// menu is created in create_screen method
//	// create gui menu
//	if (gui->CreateMenu() == -1) {
//		exit(1);
//	}
	// create global keys
	if (gui->CreateGlobalKeys() == -1) {
		exit(1);
	}

	// initialize emulation core
	emu->initialize();
	if (!emu->create_offlinesurface()) {
		exit(1);
	}
	// set again
	emu->set_window(config.window_mode, desktop_width, desktop_height);

	key_mod = emu->get_key_mod_ptr();

	// vm start
	if (emu) {
		emu->reset();
	}

	// open files
	open_recent_file();

	// main event loop
	main_event_loop();

	// release emulation core
	if(emu) {
		emu->release();
	}

	// save config
	save_config();
	release_config();

#ifdef _WIN32
	CoUninitialize();
#endif

	return 0;
}

/**
 *	quit
 */
static void exit_program(void)
{
	if (gui) {
		delete gui;
	}
	if (emu) {
		delete emu;
		emu = NULL;
	}
//	SDL_Quit();
}

/**
 *	main event loop
 */
static void main_event_loop()
{
	vm_event_loop();
}

/**
 * 	vm event loop (SDL)
 */
static void vm_event_loop()
{
	SDL_Event event;

	int total_frames = 0, draw_frames = 0, skip_frames = 0;

	int ms = 1000;
	int next_interval;
	CALC_NEXT_INTERVAL(ms, next_interval, 1000);

	int split_num = 0;

	uint32 start_time = 0;
	uint32 current_time = 0;
	int    remain_time = 0;
	uint32 next_time = 0;
	uint32 update_fps_time = next_time + 1000;

	int msgcnt = 0;
	char buf[256];

	// wait a sec.
	SDL_Delay(1000);

	// play sound
	emu->mute_sound(false);

	// process coming messages at first.
	while(SDL_PollEvent(&event)) {
		int rc = gui->ProcessEvent(&event);
		if (rc < 0) {
			working = false;
			break;
		} else if (rc > 0) {
			gui_process = false;
			gui->Show(gui_process);
		}
		vm_event_proc(&event);
	}
	gui->ProcessTimeout();

	next_time = SDL_GetTicks();
	// loop
	while(working) {
		split_num = 0;
		// get next period
		CALC_NEXT_INTERVAL(ms, next_interval, 1000);

		start_time = next_time;
//		next_time += emu->now_skip() ? 0 : next_interval;
		next_time += next_interval;
		current_time = SDL_GetTicks();

		// when vm processing is too late...
		if (next_time + 200 <= current_time) {
			// adjust next time for current time
//			next_time = current_time + (emu->now_skip() ? 0 : next_interval);
			next_time = current_time + next_interval;
		}
//		emu->out_log(LOG_DEBUG, "A1: cur:%ld st:%ld nt:%ld ms:%d int:%d", current_time, start_time, next_time, ms, next_interval);

		for(int i=0; i < (FRAME_SPLIT_NUM - 1); i++) {
			// drive machine
			if(emu)	emu->run(split_num);

			msgcnt = 10;
			while (SDL_PollEvent(&event) && msgcnt > 0) {
				int rc = gui->ProcessEvent(&event);
				if (rc < 0) {
					working = false;
					goto FIN;
					break;
				} else if (rc > 0) {
					gui_process = false;
					gui->Show(gui_process);
				}
				vm_event_proc(&event);
				msgcnt--;
			}
			gui->ProcessTimeout();

			split_num++;
		}

		if(emu) {
			// drive machine
			emu->run(split_num);

//			if (split_num + 1 == FRAME_SPLIT_NUM) {
				total_frames++;
				if(emu->now_rec_video()) {
					if (fskip_remain <= 0) {
						// rec pictures constant frames per 1 second
//						emu->mix_screen();
//						if (gui_process) {
//							gui->MixGuiScreen();
//						}
						emu->update_screen();
						draw_frames++;
						skip_frames = 0;
						fskip_remain = fskip[rec_fps_no];
					} else {
						skip_frames++;
						emu->skip_screen();
					}
					if (fskip_remain > 0) fskip_remain--;
				}
				else if(config.fps_no >= 0) {
					if (fskip_remain <= 0) {
						// constant frames per 1 second
//						emu->mix_screen();
//						if (gui_process) {
//							gui->MixGuiScreen();
//						}
						emu->update_screen();
						draw_frames++;
						skip_frames = 0;
						fskip_remain = fskip[config.fps_no];
					} else {
						skip_frames++;
						emu->skip_screen();
					}
					if (fskip_remain > 0) fskip_remain--;
				}
				else {
					current_time = SDL_GetTicks();
					if(next_time > current_time) {
						// update window if enough time
//						emu->mix_screen();
//						if (gui_process) {
//							gui->MixGuiScreen();
//						}
						emu->update_screen();
						draw_frames++;
						skip_frames = 0;
					}
					else if(++skip_frames > MAX_SKIP_FRAMES) {
						// update window at least once per 10 frames
//						emu->mix_screen();
//						if (gui_process) {
//							gui->MixGuiScreen();
//						}
						emu->update_screen();
						draw_frames++;
						skip_frames = 0;
						next_time = SDL_GetTicks();
					}
					else {
						emu->skip_screen();
					}
				}
//			}
		}
		// process event at left time in 1 frame if need
		do {
//			current_time = SDL_GetTicks();
//			remain_time = next_time - current_time;
			// check window message
			msgcnt = 100;
			while (SDL_PollEvent(&event) && msgcnt > 0) {
				int rc = gui->ProcessEvent(&event);
				if (rc < 0) {
					working = false;
					goto FIN;
					break;
				} else if (rc > 0) {
					gui_process = false;
					gui->Show(gui_process);
				}
				vm_event_proc(&event);
				msgcnt--;
			}
			gui->ProcessTimeout();

			if (skip_frames == 0) {
				current_time = SDL_GetTicks();
				remain_time = next_time - current_time;
				if (remain_time > MIN_INTERVAL) {
					if (emu) emu->update();
					current_time = SDL_GetTicks();
					remain_time = next_time - current_time;
					if (remain_time > MIN_INTERVAL) {
						SDL_Delay(MIN_INTERVAL);
						current_time = SDL_GetTicks();
						remain_time = next_time - current_time;
					}
				}
			} else {
				remain_time = 0;
			}
		} while (remain_time > 1);

//		split_num = (split_num + 1) & (FRAME_SPLIT_NUM - 1);

//		emu->out_debug("A2: rti:%ld nt:%ld (total:%d fps:%d skip:%d)",SDL_GetTicks(),next_time,total_frames,draw_frames,skip_frames);
//		emu->out_log(LOG_DEBUG,"A2: rti:%ld nt:%ld",SDL_GetTicks(),next_time);

		// calc frame rate
		current_time = SDL_GetTicks();
		if(update_fps_time <= current_time) {
			int ratio = 0;
			if (total_frames) ratio = (int)(100.0 * (double)draw_frames / (double)total_frames + 0.5);
			sprintf(buf, "%s - %d / %d fps (%d %%)", DEVICE_NAME, draw_frames, total_frames, ratio);
			SDL_WM_SetCaption(buf, NULL);

			update_fps_time += 1000;
			if(update_fps_time <= current_time) {
				update_fps_time = current_time + 1000;
			}
			total_frames = draw_frames = 0;
		}
	}

FIN:
	return;
}

/**
 *	vm event procedure (SDL)
 */
static void vm_event_proc(SDL_Event *e)
{
	switch(e->type) {
	case SDL_KEYDOWN:
	case SDL_KEYUP:
		if(emu) {
			int   type = e->key.type;
			uint16 mod = e->key.keysym.mod;
			int    sym = e->key.keysym.sym;
			emu->translate_sdl_keysym(e,&type,&sym);
			switch(type) {
			case SDL_KEYDOWN:
				emu->out_debug(_T("SDL_KEYDOWN: scancode:%02x sym:%02x(%s) mod:%02x unicode:%04x")
					,e->key.keysym.scancode,e->key.keysym.sym,SDL_GetKeyName(e->key.keysym.sym),e->key.keysym.mod,e->key.keysym.unicode);

//				if (sym != SDLK_LALT && sym != SDLK_RALT && (mod & (KMOD_LALT | KMOD_RALT))) {
				if ((*key_mod) & KEY_MOD_ALT_KEY) {
					// notify key down
					emu->system_key_down(sym);

					// execute for pressed global key
					if (emu->execute_global_keys(sym, mod)) {
						// show menu bar
//						emu->out_debug("SDL_KEYDOWN: go to gui event loop.");
//						// mute sound
//						emu->mute_sound(true);
//						// stop key repeat
//						emu->key_lost_focus();
						//
						gui_process = true;
						gui->Show(gui_process);
					}
				} else {
					emu->key_down(sym, mod);
				}
				break;
			case SDL_KEYUP:
				emu->out_debug(_T("SDL_KEYUP: scancode:%02x sym:%02x(%s) mod:%02x unicode:%04x")
					,e->key.keysym.scancode,e->key.keysym.sym,SDL_GetKeyName(e->key.keysym.sym),e->key.keysym.mod,e->key.keysym.unicode);

				emu->key_up(sym, mod);
				break;
			}
		}
		break;
//	case SDL_VIDEORESIZE:
//		break;
	case SDL_MOUSEMOTION:
		if (emu) {
			int x = ((SDL_MouseMotionEvent *)e)->x;
			int y = ((SDL_MouseMotionEvent *)e)->y;
			if (0 < y && y < 24 && 0 < x && x < SCREEN_WIDTH) {
				// show menu bar
//				emu->out_debug("SDL_MOUSEMOTION: y:%d -> Go AG ",((SDL_MouseMotionEvent *)e)->y);
//				// mute sound
//				emu->mute_sound(true);
//				// stop key repeat
//				emu->key_lost_focus();
				//
				gui_process = true;
				gui->Show(gui_process);
			}
		}
		break;
	case SDL_MOUSEBUTTONDOWN:
		if (emu) {
			SDL_MouseButtonEvent *me = (SDL_MouseButtonEvent *)e;
			// When right click at lefttop side in fullscreen, go to window mode
			if (me->button == SDL_BUTTON_RIGHT) {
				if (emu->is_fullscreen()) {
					if (me->x < 8 && me->y < 8) {
						// goto window mode
						emu->change_screen_mode(0);
					}
				}
			}
		}
		break;
	case SDL_QUIT:
		emu->out_debug(_T("SDL_QUIT"));
		working = false;
		break;
#ifdef USE_SOCKET
	case SDL_USEREVENT_SOCKET:
		if (emu) {
			int code = e->user.code;
			int no = (int)(intptr_t)e->user.data1;
			switch(code) {
				case SDL_USEREVENT_SOCKET_CONNECTED:
					emu->socket_connected(no);
					break;
				case SDL_USEREVENT_SOCKET_DISCONNECTED:
					emu->socket_disconnected(no);
					break;
				case SDL_USEREVENT_SOCKET_WRITEABLE:
					emu->socket_writeable(no);
					emu->send_data(no);
					break;
				case SDL_USEREVENT_SOCKET_READABLE:
					emu->recv_data(no);
					emu->socket_readable(no);
					break;
				case SDL_USEREVENT_SOCKET_ACCEPT:
					emu->socket_accept(no);
					break;
			}
		}
		break;
#endif
	}
}

/**
 *	get and parse options
 */
static int get_options(int ac, char *av[])
{
	int opt = 0;

	memset(app_path, 0, sizeof(app_path));
	memset(ini_path, 0, sizeof(ini_path));
	memset(ini_file, 0, sizeof(ini_file));
	memset(res_path, 0, sizeof(res_path));

	// allocate buffer
	tape_file = new char[_MAX_PATH];
	memset(tape_file, 0, sizeof(char) * _MAX_PATH);
	disk_file = new char[_MAX_PATH * 2];
	memset(disk_file, 0, sizeof(char) * _MAX_PATH * 2);
	state_file = new char[_MAX_PATH];
	memset(state_file, 0, sizeof(char) * _MAX_PATH);
	autokey_file = new char[_MAX_PATH];
	memset(autokey_file, 0, sizeof(char) * _MAX_PATH);
	reckey_file = new char[_MAX_PATH];
	memset(reckey_file, 0, sizeof(char) * _MAX_PATH);

	// set application path and name
	UTILITY::get_dir_and_basename(av[0], app_path, app_name);

	// resource path
	strncpy(res_path, app_path, _MAX_PATH);
	strncat(res_path, RESDIR, _MAX_PATH);
	strcpy(res_path, UTILITY::slim_path(res_path));

#if !defined(_WIN32) || defined(__MINGW32__)
	while ((opt = getopt(ac, av, "ha:d:i:s:t:k:")) != -1) {
		switch (opt) {
		case 'i':
			memset(ini_file, 0, sizeof(ini_file));
			strncpy(ini_file, optarg, _MAX_PATH-1);
			break;
		case 'd':
			if (disk_file[0] != '\0') {
				memset(&disk_file[_MAX_PATH], 0, sizeof(char) * _MAX_PATH);
				strncpy(&disk_file[_MAX_PATH], optarg, _MAX_PATH-1);
			} else {
				memset(&disk_file[0], 0, sizeof(char) * _MAX_PATH);
				strncpy(&disk_file[0], optarg, _MAX_PATH-1);
			}
			break;
		case 't':
			memset(tape_file, 0, sizeof(char) * _MAX_PATH);
			strncpy(tape_file, optarg, _MAX_PATH-1);
			break;
		case 's':
			memset(state_file, 0, sizeof(char) * _MAX_PATH);
			strncpy(state_file, optarg, _MAX_PATH-1);
			break;
		case 'a':
			memset(autokey_file, 0, sizeof(char) * _MAX_PATH);
			strncpy(autokey_file, optarg, _MAX_PATH-1);
			break;
		case 'k':
			memset(reckey_file, 0, sizeof(char) * _MAX_PATH);
			_tcsncpy(reckey_file, optarg, _MAX_PATH-1);
			break;
		case 'h':
		case '?':
			printf("Usage: %s [-h] [-i <ini_file>] [-t <tape_file>] [-d <disk_file>] [-s <state_file>] [-a <autokey_file>] [<support_file> ...]\n",app_name);
			exit(0);
			break;
		default:
			break;
		}
	}
#else
	int optind = 1;
#endif
	for (;optind < ac && av[optind][0] != '\0'; optind++) {
		int rc = check_supported_file(av[optind]);
		switch(rc) {
		case 1:
			memset(tape_file, 0, sizeof(char) * _MAX_PATH);
			strncpy(tape_file, av[optind], _MAX_PATH - 1);
			break;
		case 2:
			if (disk_file[0] != _T('\0')) {
				memset(&disk_file[_MAX_PATH], 0, sizeof(char) * _MAX_PATH);
				strncpy(&disk_file[_MAX_PATH], av[optind], _MAX_PATH - 1);
			} else {
				memset(&disk_file[0], 0, sizeof(char) * _MAX_PATH);
				strncpy(&disk_file[0], av[optind], _MAX_PATH - 1);
			}
			break;
		case 3:
			memset(state_file, 0, sizeof(char) * _MAX_PATH);
			strncpy(state_file, av[optind], _MAX_PATH - 1);
			break;
		case 4:
			memset(autokey_file, 0, sizeof(char) * _MAX_PATH);
			strncpy(autokey_file, av[optind], _MAX_PATH - 1);
			break;
		case 5:
			memset(ini_file, 0, sizeof(ini_file));
			strncpy(ini_file, av[optind], _MAX_PATH - 1);
			break;
		case 6:
			memset(reckey_file, 0, sizeof(char) * _MAX_PATH);
			strncpy(reckey_file, av[optind], _MAX_PATH - 1);
			break;
		default:
			break;
		}
	}

	// check ini file
	if (ini_file[0] != _T('\0')) {
		if (UTILITY::check_file_extension(ini_file, _T("ini"))) {
			_TCHAR ini_name[_MAX_PATH];
			UTILITY::get_dir_and_basename(ini_file, ini_path, ini_name);
		}
		UTILITY::add_path_separator(ini_path);
	} else {
		strncpy(ini_path, app_path, _MAX_PATH);
		strncpy(ini_file, ini_path, _MAX_PATH);
		strncat(ini_file, _T(CONFIG_NAME), _MAX_PATH);
		strncat(ini_file, _T(".ini"), _MAX_PATH);
	}

	return 0;
}

/**
 *	open recent file
 */
static void open_recent_file()
{
	char path[_MAX_PATH];

	if (emu) {
		if (tape_file[0] != '\0') {
			strcpy(path, tape_file);
			gui->LoadDataRec(path);
		}
#ifdef USE_FD1
		// auto open recent file
		for(int drv=0; drv<2; drv++) {
			path[0] = '\0';
			int bank_num = 0;
			if (disk_file[drv * _MAX_PATH] != '\0') {
				strcpy(path, &disk_file[drv * _MAX_PATH]);
			}
			else if (_tcslen(config.recent_disk_path[drv][0].path) > 0) {
				strcpy(path, config.recent_disk_path[drv][0].path);
				bank_num = config.recent_disk_path[drv][0].num;
			}
			if (path[0] != '\0') {
				gui->OpenFloppyDisk(drv, path, bank_num, 0, false);
			}
		}
#endif
		if (state_file[0] != '\0') {
			strcpy(path, state_file);
			gui->LoadStatus(path);
		}
		if (autokey_file[0] != '\0') {
			strcpy(path, autokey_file);
			gui->LoadAutoKey(path);
		}
		if (reckey_file[0] != '\0') {
			_tcscpy(path, reckey_file);
			gui->LoadRecKey(path);
		}
	}

	// release buffer
	delete[] tape_file; tape_file = NULL;
	delete[] disk_file; disk_file = NULL;
	delete[] state_file; state_file = NULL;
	delete[] autokey_file; autokey_file = NULL;
	delete[] reckey_file; reckey_file = NULL;
}

/**
 *	check supported file
 */
int check_supported_file(char *file_path)
{
	bool rc;

	rc = UTILITY::check_file_extension(file_path, _T(".l3"))
	  || UTILITY::check_file_extension(file_path, _T(".l3b"))
	  || UTILITY::check_file_extension(file_path, _T(".l3c"))
	  || UTILITY::check_file_extension(file_path, _T(".wav"))
	  || UTILITY::check_file_extension(file_path, _T(".t9x"));
	if (rc) {
		// tape image
		return 1;
	}

#ifdef USE_FD1
	rc = UTILITY::check_file_extension(file_path, _T(".d88"))
	  || UTILITY::check_file_extension(file_path, _T(".td0"))
	  || UTILITY::check_file_extension(file_path, _T(".imd"))
	  || UTILITY::check_file_extension(file_path, _T(".dsk"))
	  || UTILITY::check_file_extension(file_path, _T(".fdi"))
	  || UTILITY::check_file_extension(file_path, _T(".hdm"))
	  || UTILITY::check_file_extension(file_path, _T(".tfd"))
	  || UTILITY::check_file_extension(file_path, _T(".xdf"))
	  || UTILITY::check_file_extension(file_path, _T(".2d"))
	  || UTILITY::check_file_extension(file_path, _T(".sf7"));
	if (rc) {
		// disk image
		return 2;
	}
#endif

#ifdef _BML3MK5
	rc = UTILITY::check_file_extension(file_path, _T(".l3r"));
	if (rc) {
		// maybe state file
		return 3;
	}

	rc = UTILITY::check_file_extension(file_path, _T(".txt"))
	  || UTILITY::check_file_extension(file_path, _T(".bas"))
	  || UTILITY::check_file_extension(file_path, _T(".lpt"));
	if (rc) {
		// maybe text file (for autokey)
		return 4;
	}

	rc = UTILITY::check_file_extension(file_path, _T(".ini"));
	if (rc) {
		// maybe ini file
		return 5;
	}

	rc = UTILITY::check_file_extension(file_path, _T(".l3k"));
	if (rc) {
		// maybe record key file
		return 6;
	}
#endif

	return 0;
}
