/** @file sdl_screen.cpp

	Skelton for retropc emulator
	SDL edition

	@author Sasaji
	@date   2012.02.21

	@brief [ sdl screen ]

	@note
	This code is based on win32_screen.cpp of the Common Source Code Project.
	Author : Takeda.Toshiya
*/

#include "emu.h"
#include "vm/vm.h"
#include "config.h"
#include "gui/ag_gui.h"

static const Size window_proposal[] = {
	{640,400},{640,480},{800,600},{1024,768},{0,0}
};

void EMU::EMU_SCREEN()
{
	screen_width = SCREEN_WIDTH;
	screen_height = SCREEN_HEIGHT;
	screen_width_aspect = SCREEN_WIDTH_ASPECT;
	screen_height_aspect = SCREEN_HEIGHT_ASPECT;
	window_width = WINDOW_WIDTH;
	window_height = WINDOW_HEIGHT;
	min_window_width = MIN_WINDOW_WIDTH;
	min_window_height = MIN_WINDOW_HEIGHT;
	screen_size_changed = true;
	desktop_width = MIN_WINDOW_WIDTH;
	desktop_height = MIN_WINDOW_HEIGHT;

	source_width = screen_width;
	source_height = screen_height;
	source_width_aspect = source_height_aspect = -1;
	stretch_pow_x = stretch_pow_y = -1;
	display_stretched_width = display_stretched_height = 1;

	// screen mode
	screen_mode_count = 0;
	window_mode_count = 0;
	window_mode_power = 10;
	now_fullscreen = false;
	prev_window_mode = 0;

	rec_video_size[0].w = MIN_WINDOW_WIDTH;
	rec_video_size[0].h = MIN_WINDOW_HEIGHT;
	rec_video_size[0].x = (MAX_WINDOW_WIDTH - MIN_WINDOW_WIDTH) / 2;
	rec_video_size[0].y = (MAX_WINDOW_HEIGHT - MIN_WINDOW_HEIGHT) / 2;
	rec_video_size[1].w = MAX_WINDOW_WIDTH;
	rec_video_size[1].h = MAX_WINDOW_HEIGHT;
	rec_video_size[1].x = 0;
	rec_video_size[1].y = 0;

	// initialize video recording
	now_rec_vid = false;

	// initialize update flags
	first_draw_screen = false;
	first_invalidate = self_invalidate = false;
	skip_frame = false;

	//
	screen = NULL;
	srcsurface = NULL;
	mixsurface = NULL;
	screen_flags = 0;

#ifdef _BML3MK5
	// create ledbox
	ledbox = new LedBox();
	msgboard = new MsgBoard(this);
#endif
}

void EMU::initialize_screen()
{
	out_debug(_T("EMU::initialize_screen"));


	if (msgboard) {
		msgboard->InitScreen(&pixel_format, screen_width, screen_height);
		msgboard->SetMessagePos(4 + reMix.x, -4 - reMix.y, 2);
		msgboard->SetInfoPos(-4 - reMix.x, 4 + reMix.y, 1);
		msgboard->SetVisible(config.show_dlg & 2 ? true : false);
	}
	if (ledbox) {
		if (!ledbox->Create(res_path, &pixel_format)) {
			delete ledbox;
			ledbox = NULL;
		}
	}

#ifdef USE_OPENGL
	initialize_opengl();
#endif
}

void EMU::release_screen()
{
	// stop video recording
	stop_rec_video();

	if (srcsurface != NULL) SDL_FreeSurface(srcsurface);
	if (mixsurface != NULL) SDL_FreeSurface(mixsurface);

#ifdef USE_OPENGL
	release_opengl();
#endif
}

bool EMU::create_screen(int width, int height, Uint32 flags)
{
	if (now_fullscreen) {
		flags |= SDL_FULLSCREEN;
	} else {
		flags &= ~SDL_FULLSCREEN;
	}
	flags &= ~SDL_OPENGL;
#ifdef USE_OPENGL
	if (use_opengl) {
		flags |= SDL_OPENGL;
		set_opengl_attr();
	}
#endif
	screen = SDL_SetVideoMode(width, height, _RGB888, flags);
	if (screen == NULL) {
		out_log(LOG_ERROR, _T("SDL_SetVideoMode: %s."), SDL_GetError());
		return false;
	}
	if (flags != screen->flags) {
		out_log(LOG_WARN, _T("SDL_SetVideoMode: different flags %x -> %x."), flags, screen->flags);
	}
	screen_flags = screen->flags;
	pixel_format = *(screen->format);

#ifdef _DEBUG_LOG
	out_debug(_T("screen->flags:%x"),screen->flags);
	out_debug(_T("screen->w:%d h:%d"),screen->w, screen->h);
	out_debug(_T("screen->pitch:%d"), screen->pitch);
	out_debug(_T("screen->format->BitsPerPixel:%d"),screen->format->BitsPerPixel);
	out_debug(_T("screen->format->BytesPerPixel:%d"),screen->format->BytesPerPixel);
	out_debug(_T("screen->format->Rmask:%x, Gmask:%x, Bmask:%x, Amask:%x"),screen->format->Rmask,screen->format->Gmask,screen->format->Bmask,screen->format->Amask);
	out_debug(_T("screen->format->Rshift:%d, Gshift:%d, Bshift:%d, Ashift:%d"),screen->format->Rshift,screen->format->Gshift,screen->format->Bshift,screen->format->Ashift);
	out_debug(_T("screen->format->Rloss:%d, Gloss:%d, Bloss:%d, Aloss:%d"),screen->format->Rloss,screen->format->Gloss,screen->format->Bloss,screen->format->Aloss);
	out_debug(_T("screen->format->colorkey:%d"),screen->format->colorkey);
	out_debug(_T("screen->format->alpha:%d"),screen->format->alpha);
#endif
#ifdef USE_OPENGL
	if (flags & SDL_OPENGL) {
		GLubyte *vender   = (GLubyte *)glGetString(GL_VENDOR);
		GLubyte *renderer = (GLubyte *)glGetString(GL_RENDERER);
		GLubyte *version  = (GLubyte *)glGetString(GL_VERSION);
		out_log(LOG_DEBUG, _T("Use OPENGL: VENDER:%s RENDERER:%s VERSION:%s"), vender, renderer, version);

		// set pixel format
		set_pixel_format();

		// setting polygon
		set_opengl_poly(screen->w, screen->h);

	} else {
		use_opengl = 0;
	}
#endif
	// create gui window
	if (gui->InitVideoSDL(screen, width, height) == -1) {
		return false;
	}
	return true;
}

bool EMU::create_offlinesurface()
{
	if (screen != NULL) {
		srcsurface = SDL_CreateRGBSurface(SDL_SWSURFACE, screen_width, screen_height, pixel_format.BitsPerPixel,
			pixel_format.Rmask, pixel_format.Gmask, pixel_format.Bmask, pixel_format.Amask);
		if (srcsurface == NULL) {
			out_log(LOG_ERROR, _T("SDL_CreateRGBSurface: srcsurface: %s."),SDL_GetError());
			return false;
		}
		// mixsurface size normalize 2^n for opengl
		int mix_width = 0;
		int mix_height = 0;
		for(int i=16; i>=4; i--) {
			if (screen_width <= (1 << i)) mix_width = (1 << i);
			if (screen_height <= (1 << i)) mix_height = (1 << i);
		}
		mix_s = (double)screen_width / mix_width;
		mix_t = (double)screen_height / mix_height;
		mixsurface = SDL_CreateRGBSurface(SDL_SWSURFACE, mix_width, mix_height, pixel_format.BitsPerPixel,
			pixel_format.Rmask, pixel_format.Gmask, pixel_format.Bmask, pixel_format.Amask);
		if (mixsurface == NULL) {
			out_log(LOG_ERROR, _T("SDL_CreateRGBSurface: mixsurface: %s."),SDL_GetError());
			return false;
		}
	}
	return true;
}

void EMU::get_rgbformat(uint32 *r_mask, uint32 *g_mask, uint32 *b_mask, uint8 *r_shift, uint8 *g_shift, uint8 *b_shift)
{
	if (screen) {
		*r_mask = pixel_format.Rmask;
		*g_mask = pixel_format.Gmask;
		*b_mask = pixel_format.Bmask;
		*r_shift = pixel_format.Rshift;
		*g_shift = pixel_format.Gshift;
		*b_shift = pixel_format.Bshift;
	} else {
		*r_mask = 0;
		*g_mask = 0;
		*b_mask = 0;
		*r_shift = 0;
		*g_shift = 0;
		*b_shift = 0;
	}
}

void EMU::get_rgbcolor(scrntype pixel, uint8 *r, uint8 *g, uint8 *b)
{
	if (screen) {
		SDL_GetRGB((Uint32)pixel, &pixel_format, r, g, b);
	} else {
		*r = 0;
		*g = 0;
		*b = 0;
	}
}

scrntype EMU::map_rgbcolor(uint8 r, uint8 g, uint8 b)
{
	if (screen) {
		return (scrntype)SDL_MapRGB(&pixel_format, r, g, b);
	} else {
		return 0;
	}
}

int EMU::get_window_width(int mode)
{
	return window_width + screen_width_aspect * mode;
}

int EMU::get_window_height(int mode)
{
	return window_height + screen_height_aspect * mode;
}

void EMU::set_display_size(int width, int height, int power, bool window_mode)
{
	bool stretch_changed = false;

	if(width != -1 && (display_width != width || display_height != height)) {
		display_width = width;
		display_height = height;
		stretch_changed = true;
	}

	int source_width_aspect, source_height_aspect;

	source_width = screen_width;
	source_height = screen_height;
	source_width_aspect = screen_width_aspect;
	source_height_aspect = screen_height_aspect;

	if(config.stretch_screen && !window_mode && display_width >= source_width && display_height >= source_height) {
		if (!config.cutout_screen) {
			// fit to full screen
			stretched_width = (display_height * source_width_aspect) / source_height_aspect;
			stretched_height = display_height;
			if(stretched_width > display_width) {
				stretched_width = display_width;
				stretched_height = (display_width * source_height_aspect) / source_width_aspect;
			}
			display_stretched_width = display_width;
			display_stretched_height = display_height;

			screen_dest_x = (display_width - stretched_width) / 2;
			screen_dest_y = (display_height - stretched_height) / 2;
			screen_stretched_dest_x = screen_dest_x;
			screen_stretched_dest_y = screen_dest_y;

		} else {
			// fit text area to full screen
			stretched_width = (display_height * min_window_width) / min_window_height * screen_width / min_window_width;
			stretched_height = display_height * screen_height / min_window_height;
			if(stretched_width > display_width * screen_width / min_window_width) {
				stretched_width = display_width * screen_width / min_window_width;
				stretched_height = (display_width * min_window_height) / min_window_width * screen_height / min_window_height;
			}
			display_stretched_width = min_window_width;
			display_stretched_height = min_window_height;

			screen_dest_x = (display_width - stretched_width) / 2;
			screen_dest_y = (display_height - stretched_height) / 2;
			screen_stretched_dest_x = (min_window_width - screen_width) / 2;
			screen_stretched_dest_y = (min_window_height - screen_height) / 2;

		}
	}
	else {
		stretched_width = source_width_aspect * power / 10;
		stretched_height = source_height_aspect * power / 10;
		display_stretched_width = display_width * 10 / power;
		display_stretched_height = display_height * 10 / power;

		screen_dest_x = (display_width - stretched_width) / 2;
		screen_dest_y = (display_height - stretched_height) / 2;
		screen_stretched_dest_x = screen_dest_x * 10 / power;
		screen_stretched_dest_y = screen_dest_y * 10 / power;
	}

	reMix.x = 0; reMix.y = 0; reMix.w = source_width; reMix.h = source_height;
	reSuf.x = screen_dest_x; reSuf.y = screen_dest_y; reSuf.w = stretched_width; reSuf.h = stretched_height;
#ifdef USE_OPENGL
	reTex = reSuf;
#endif
	if (screen_dest_x < 0) {
		reMix.x = -screen_stretched_dest_x; reMix.w = display_stretched_width;
		reSuf.x = 0;						reSuf.w = display_width;
	}
	if (screen_dest_y < 0) {
		reMix.y = -screen_stretched_dest_y; reMix.h = display_stretched_height;
		reSuf.y = 0;						reSuf.h = display_height;
	}

	// send display size to vm
	set_vm_display_size();

	first_draw_screen = false;
	first_invalidate = true;

#ifdef _BML3MK5
	if (ledbox) {
		ledbox->SetPos(reMix.x, reMix.y, reMix.x+reMix.w, reMix.y+reMix.h, config.led_pos);
	}
	if (msgboard) {
		msgboard->SetSize(source_width, source_height);
		msgboard->SetMessagePos(4 + reMix.x, -4 - reMix.y, 2);
		msgboard->SetInfoPos(-4 - reMix.x, 4 + reMix.y, 1);
	}
#endif

}

void EMU::change_screen_size(int sw, int sh, int swa, int sha, int ww, int wh)
{
}

void EMU::set_vm_display_size()
{
	if (vm) {
		int new_size = config.screen_video_size;

		if (rec_video_size[new_size].w > reMix.w
			|| rec_video_size[new_size].h > reMix.h) {
			vm->set_display_size(rec_video_size[new_size].x, rec_video_size[new_size].y,
				rec_video_size[new_size].x + rec_video_size[new_size].w, rec_video_size[new_size].y + rec_video_size[new_size].h);
		} else {
			vm->set_display_size(reMix.x,reMix.y,reMix.x+reMix.w,reMix.y+reMix.h);
		}
	}
}

void EMU::draw_screen()
{
	// draw screen
	SDL_LockSurface(srcsurface);
	vm->draw_screen();
	SDL_UnlockSurface(srcsurface);

	first_draw_screen = true;
}

void EMU::mix_screen()
{
	SDL_BlitSurface(srcsurface, NULL, mixsurface, NULL);
	if (ledbox && (config.show_dlg & 1)) {
		ledbox->Update(mixsurface, update_led());
	}
	if (msgboard && (config.show_dlg & 2)) {
		msgboard->Draw(mixsurface);
	}

#ifdef USE_OPENGL
	if (screen_flags & SDL_OPENGL) {
		GLenum gl_errno;

		// draw texture using screen pixel buffer
		glBindTexture(GL_TEXTURE_2D, mix_texture_name);

		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, mixsurface->w, mixsurface->h, 0, GL_RGBA,
				GL_UNSIGNED_BYTE, (GLvoid *)mixsurface->pixels);
		if ((gl_errno = glGetError()) != GL_NO_ERROR) {
			out_log(LOG_ERROR, _T("glTexImage2D: %d"), gl_errno);
		}

		// fill black
		glBegin(GL_QUADS);
			glColor3ub(0, 0, 0);
			glVertex3i(0, 0, 0);
			glVertex3i(0, display_height, 0);
			glVertex3i(display_width, display_height, 0);
			glVertex3i(display_width, 0, 0);
		glEnd();

		glEnable(GL_TEXTURE_2D);
		glColor3ub(255, 255, 255);
		glBegin(GL_QUADS);
			// anticlockwise
			glTexCoord2d(0.0, 0.0);
			glVertex3i(reTex.x, reTex.y, 0);
			glTexCoord2d(0.0, mix_t);
			glVertex3i(reTex.x, reTex.y + reTex.h, 0);
			glTexCoord2d(mix_s, mix_t);
			glVertex3i(reTex.x + reTex.w, reTex.y + reTex.h, 0);
			glTexCoord2d(mix_s, 0.0);
			glVertex3i(reTex.x + reTex.w, reTex.y, 0);
		glEnd();
		glDisable(GL_TEXTURE_2D);

		glBindTexture(GL_TEXTURE_2D, 0);

	} else
#endif
	{
		if (window_mode_power > 10 || (now_fullscreen && config.stretch_screen)) {
			zoomin_surface(mixsurface, &reMix, screen, &reSuf);
		} else {
			SDL_BlitSurface(mixsurface, &reMix, screen, &reSuf);
		}
	}

	// invalidate window

	self_invalidate = true;
	skip_frame = false;

	// record picture
//	if(now_rec_vid) {
//		if(AVIStreamWrite(pAVICompressed, rec_frames++, 1, (LPBYTE)lpBmpSource, pbmInfoHeader->biSizeImage, AVIIF_KEYFRAME, NULL, NULL) != AVIERR_OK) {
//			stop_rec_video();
//		}
//	}
}

scrntype* EMU::screen_buffer(int y)
{
	return (scrntype *)srcsurface->pixels + screen_width * y;
}

int EMU::screen_buffer_offset()
{
	return screen_width;
}

void EMU::update_screen()
{
	mix_screen();

#ifdef USE_OPENGL
	if (screen_flags & SDL_OPENGL) {
		gui->MixGuiScreen();

		SDL_GL_SwapBuffers();
	} else
#endif
	{
		// GuiMixScreen include SDL_UpdateRect process.
		if (!gui->MixGuiScreen()) {
			SDL_UpdateRect(screen, 0, 0, display_width, display_height);
		}
	}
}

void EMU::copy_surface(SDL_Surface *src, SDL_Rect *reSrc, SDL_Surface *dst, SDL_Rect *reDst)
{
	scrntype *pSrc;
	scrntype *pDst;
	Uint8 r, g, b, a;

	int sright = (reSrc->x + reSrc->w);
	int sbottom = (reSrc->y + reSrc->h);
	int dright = (reDst->x + reDst->w);
	int dbottom = (reDst->y + reDst->h);

	SDL_LockSurface(dst);
	SDL_LockSurface(src);
	for(int sy=reSrc->y, dy = reDst->y; sy < sbottom && dy < dbottom; sy++, dy++) {
		pSrc = (scrntype *)src->pixels + sy * (src->w);
		pDst = (scrntype *)dst->pixels + dy * (dst->w);
		for(int sx=reSrc->x, dx = reDst->x; sx < sright && dx < dright; sx++, dx++) {
			r = ((pSrc[sx] & src->format->Rmask) >> (src->format->Rshift)) & 0xff;
			g = ((pSrc[sx] & src->format->Gmask) >> (src->format->Gshift)) & 0xff;
			b = ((pSrc[sx] & src->format->Bmask) >> (src->format->Bshift)) & 0xff;
			a = ((pSrc[sx] & src->format->Amask) >> (src->format->Ashift)) & 0xff;
			pDst[dx] =
				((r << dst->format->Rshift) & dst->format->Rmask)
				| ((g << dst->format->Gshift) & dst->format->Gmask)
				| ((b << dst->format->Bshift) & dst->format->Bmask)
				| ((a << dst->format->Ashift) & dst->format->Amask);
		}
	}
	SDL_UnlockSurface(src);
	SDL_UnlockSurface(dst);
}

void EMU::zoomin_surface(SDL_Surface *src, SDL_Rect *reSrc, SDL_Surface *dst, SDL_Rect *reDst)
{
	scrntype *pSrc;
	scrntype *pDst;

	int sright = (reSrc->x + reSrc->w);
	int sbottom = (reSrc->y + reSrc->h);
	int dright = (reDst->x + reDst->w);
	int dbottom = (reDst->y + reDst->h);
	int x_mag = (reSrc->w << 12) / reDst->w;
	int y_mag = (reSrc->h << 12) / reDst->h;
	int sx_m = 0;
	int sy_m = 0;

	SDL_LockSurface(dst);
	SDL_LockSurface(src);
	sy_m = (reSrc->y << 12);
	for(int sy=reSrc->y, dy = reDst->y; sy < sbottom && dy < dbottom; dy++) {
		pSrc = (scrntype *)src->pixels + sy * (src->w);
		pDst = (scrntype *)dst->pixels + dy * (dst->w);
		sx_m = (reSrc->x << 12);
		for(int sx=reSrc->x, dx = reDst->x; sx < sright && dx < dright; dx++) {
			pDst[dx] = pSrc[sx];
			sx_m+=x_mag;
			sx = (sx_m >> 12);
		}
		sy_m+=y_mag;
		sy = (sy_m >> 12);
	}
	SDL_UnlockSurface(src);
	SDL_UnlockSurface(dst);
}

void EMU::capture_screen()
{
	const _TCHAR *app_path;
	_TCHAR file_path[_MAX_PATH];
	app_path = _tcslen(config.snapshot_path) > 0 ? config.snapshot_path : application_path();
	_stprintf(file_path, _T("%s%04d-%02d-%02d_%02d-%02d-%02d.bmp"), app_path
		, sTime.tm_year + 1900, sTime.tm_mon + 1, sTime.tm_mday, sTime.tm_hour, sTime.tm_min, sTime.tm_sec);

	SDL_Surface *bmpsurface = NULL;
	int size = config.screen_video_size;
	// notice that format is same as screen because SaveBMP cannot out pixeldata
	bmpsurface = SDL_CreateRGBSurface(SDL_SWSURFACE, rec_video_size[size].w, rec_video_size[size].h
		, screen->format->BitsPerPixel, screen->format->Rmask, screen->format->Gmask, screen->format->Bmask, screen->format->Amask);

	if (bmpsurface == NULL) {
		out_log(LOG_ERROR, _T("Cannot create screen captured surface: %s"), SDL_GetError());
		return;
	}

	// copy surface
	SDL_Rect re = { 0, 0, rec_video_size[size].w, rec_video_size[size].h };

	// In OpenGL mode, SDL_BlitSurface does not copy surface....
	copy_surface(mixsurface, &rec_video_size[size], bmpsurface, &re);

	// save surface
	if (SDL_SaveBMP(bmpsurface, file_path) < 0) {
		out_log(LOG_ERROR, _T("Cannot save screen captured file: %s"), SDL_GetError());
	}
	SDL_FreeSurface(bmpsurface);
}

void EMU::start_rec_video(int fps, bool show_dialog)
{
}

void EMU::stop_rec_video()
{
}

void EMU::restart_rec_video()
{
}

void EMU::resize_rec_video(int new_size)
{
	config.screen_video_size = new_size;
}

//----------

int EMU::qsort_by_w_h_n(const void *a, const void *b)
{
	struct window_mode_st *a1 = (struct window_mode_st *)a;
	struct window_mode_st *b1 = (struct window_mode_st *)b;

	int rc = 0;
	if (!rc) rc = (a1->width - b1->width);
	if (!rc) rc = (a1->height - b1->height);
	if (!rc) rc = (a1->power - b1->power);
	return rc;
}

int EMU::qsort_by_n_w_h(const void *a, const void *b)
{
	struct screen_mode_st *a1 = (struct screen_mode_st *)a;
	struct screen_mode_st *b1 = (struct screen_mode_st *)b;

	int rc = 0;
	if (!rc) rc = (a1->disp_no - b1->disp_no);
	if (!rc) rc = (a1->width - b1->width);
	if (!rc) rc = (a1->height - b1->height);
	return rc;
}

void EMU::enum_window_mode(int max_width, int max_height)
{
	// enumerate screen mode for window
	int w,h;
	bool minsize = false;
	bool maxsize = false;

	window_mode_count = 0;
	for (int i=0; window_proposal[i].cx != 0 && window_mode_count < WINDOW_MODE_MAX; i++) {
		w = window_proposal[i].cx;
		h = window_proposal[i].cy;
		// minimum window size?
		if (w < MIN_WINDOW_WIDTH || h < MIN_WINDOW_HEIGHT) {
			continue;
		}
		if (!minsize) {
			w = MIN_WINDOW_WIDTH;
			h = MIN_WINDOW_HEIGHT;
			i--;
			minsize = true;
		}
		if (w > MAX_WINDOW_WIDTH || h > MAX_WINDOW_HEIGHT) {
			if (!maxsize) {
				w = MAX_WINDOW_WIDTH;
				h = MAX_WINDOW_HEIGHT;
				maxsize = true;
				i--;
			} else {
				break;
			}
		}
		for (int pow = 10; pow <= 30 && window_mode_count < WINDOW_MODE_MAX; pow+=5) {
			if ((w*pow/10) <= max_width && (h*pow/10) <= max_height) {
				bool found = false;
				for(int j = 0; j < window_mode_count; j++) {
					if(window_mode[j].power == pow && window_mode[j].width == (w*pow/10) && window_mode[j].height == (h*pow/10)) {
						found = true;
						break;
					}
				}
				if (!found) {
					window_mode[window_mode_count].power = pow;
					window_mode[window_mode_count].width = w * pow / 10;
					window_mode[window_mode_count].height = h * pow / 10;
					out_debug(_T("window_mode:%d %dx%d x%.1f"),window_mode_count, w * pow / 10, h * pow / 10, (double)pow / 10.0);
					window_mode_count++;
				}
			} else {
				break;
			}
		}
	}
	qsort(&window_mode, window_mode_count, sizeof(struct window_mode_st), qsort_by_w_h_n);
}

void EMU::enum_screen_mode(SDL_PixelFormat *format, Uint32 flags)
{
	// enumerate screen mode for fullscreen
	SDL_Rect **list = NULL;
	int disp_no = 0;
	screen_mode_count = 0;
	flags |= SDL_FULLSCREEN;

	list = SDL_ListModes(format, flags);

	if (list != NULL && list != (SDL_Rect **)-1) {
		for(int i = 0; list[i] != NULL && screen_mode_count < SCREEN_MODE_MAX; i++) {
			if(list[i]->w < MIN_WINDOW_WIDTH || list[i]->h < MIN_WINDOW_HEIGHT || list[i]->w < 640 || list[i]->h < 480) {
				out_debug(_T("screen_mode:-- [%d] %dx%d ignored"), disp_no, list[i]->w, list[i]->h);
				continue;
			}
			if(((list[i]->w * 30 / list[i]->h) != 40)
			&& ((list[i]->w * 40 / list[i]->h) != 50)
			&& ((list[i]->w * 90 / list[i]->h) != 160)
			&& ((list[i]->w * 50 / list[i]->h) != 80)) {
				out_debug(_T("screen_mode:-- [%d] %dx%d ignored"), disp_no, list[i]->w, list[i]->h);
				continue;
			}
			if (list[i]->w <= 1280 && (list[i]->w != 640 && list[i]->w != 800 && list[i]->w != 1024 && list[i]->w != 1280)) {
				out_debug(_T("screen_mode:-- [%d] %dx%d ignored"), disp_no, list[i]->w, list[i]->h);
				continue;
			}
			bool found = false;
			for(int j = 0; j < screen_mode_count; j++) {
				if(screen_mode[j].disp_no == disp_no && screen_mode[j].width == list[i]->w && screen_mode[j].height == list[i]->h) {
					found = true;
					out_debug(_T("screen_mode:-- [%d] %dx%d already exist"), disp_no, list[i]->w, list[i]->h);
					break;
				}
			}
			if(!found) {
				screen_mode[screen_mode_count].disp_no = disp_no;
				screen_mode[screen_mode_count].width = list[i]->w;
				screen_mode[screen_mode_count].height = list[i]->h;
				out_debug(_T("screen_mode:%2d [%d] %dx%d"), screen_mode_count, disp_no, list[i]->w, list[i]->h);
				screen_mode_count++;
			}
		}
	}
	qsort(&screen_mode, screen_mode_count, sizeof(struct screen_mode_st), qsort_by_n_w_h);
}

void EMU::find_screen_mode()
{
	if (config.window_mode < 8) {
		bool find = false;
		int i = config.window_mode;
		if (i < window_mode_count) {
			if (window_mode[i].width == config.screen_width
			&&  window_mode[i].height == config.screen_height) {
				find = true;
				config.window_mode = i;
			}
		}
		if (!find) {
			config.window_mode = 0;
		}
	} else {
		bool find = false;
		for(int i = 0; i < screen_mode_count; i++) {
			if (screen_mode[i].disp_no == config.disp_device_no
			&&  screen_mode[i].width == config.screen_width
			&&  screen_mode[i].height == config.screen_height) {
				find = true;
				config.window_mode = i + 8;
				break;
			}
		}
		if (!find) {
			config.window_mode = 0;
		}
	}
}

void EMU::change_screen_mode(int mode)
{
	if (mode == config.window_mode) return;

	if (mode == -1) {
		if (now_fullscreen) {
			mode = prev_window_mode;
		}
	}
	if (!now_fullscreen) {
		prev_window_mode = config.window_mode;
	}

	set_window(mode, desktop_width, desktop_height);
	if (!create_screen(config.screen_width, config.screen_height, screen_flags)) {
		exit(1);
	}

}

void EMU::change_stretch_screen()
{
	config.stretch_screen = !config.stretch_screen;
	if (now_fullscreen) set_display_size(-1, -1, 10, !now_fullscreen);
}

void EMU::change_cutout_screen()
{
	config.cutout_screen = !config.cutout_screen;
	if (now_fullscreen) set_display_size(-1, -1, 10, !now_fullscreen);
}

void EMU::set_window(int mode, int cur_width, int cur_height)
{
	if(mode >= 0 && mode < 8) {
		// go window
		if (mode >= window_mode_count) mode = 0;
		int width = window_mode[mode].width;
		int height = window_mode[mode].height;

		config.window_mode = mode;
		config.disp_device_no = 0;
		config.screen_width = width;
		config.screen_height = height;
		window_mode_power = window_mode[mode].power;

		now_fullscreen = false;

		// set screen size to emu class
		set_display_size(width, height, window_mode_power, true);
	}
	else if(!now_fullscreen) {
		// go fullscreen
		if (mode >= screen_mode_count + 8) mode = -1;
		int disp_no = (mode == -1) ? 0 : screen_mode[mode - 8].disp_no;
		int width = (mode == -1) ? cur_width : screen_mode[mode - 8].width;
		int height = (mode == -1) ? cur_height : screen_mode[mode - 8].height;

		if(mode == -1) {
			for(int i = 0; i < screen_mode_count; i++) {
				if(screen_mode[i].disp_no == disp_no && screen_mode[i].width == width && screen_mode[i].height == height) {
					mode = i + 8;
					break;
				}
			}
		}
		config.window_mode = mode;
		config.disp_device_no = disp_no;
		config.screen_width = width;
		config.screen_height = height;

		now_fullscreen = true;

		// set screen size to emu class
		set_display_size(width, height, 10, false);
	}
}

bool EMU::get_window_mode_str(int num, char *str)
{
	if (num < window_mode_count) {
		sprintf(str, "  %dx%d x%.1f",
			window_mode[num].width,
			window_mode[num].height,
			(double)window_mode[num].power / 10.0);
		return true;
	} else {
		sprintf(str, "  ");
	}
	return false;
}

bool EMU::get_screen_mode_str(int num, char *str)
{
	if (num < screen_mode_count) {
		sprintf(str, "  %dx%d",
			screen_mode[num].width,
			screen_mode[num].height);
		return true;
	} else {
		sprintf(str, "  ");
	}
	return false;
}

bool EMU::get_screen_record_size_str(int num, char *str)
{
	if (0 <= num && num <= 1) {
		sprintf(str, "  %dx%d",
			rec_video_size[num].w,
			rec_video_size[num].h);
		return true;
	} else {
		sprintf(str, "  ");
	}
	return false;
}

// ----------
#ifdef USE_OPENGL

void EMU::initialize_opengl()
{
	if (!use_opengl) {
		return;
	}

	GLenum gl_errno;

	mix_texture_name = 0;

	// create
#if defined(_RGB555) || defined(_RGB565)
	glPixelStorei(GL_UNPACK_ALIGNMENT, 2);
#elif defined(_RGB888)
	glPixelStorei(GL_UNPACK_ALIGNMENT, 4);
#else
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
#endif

	glGenTextures(1, &mix_texture_name);
	if ((gl_errno = glGetError()) != GL_NO_ERROR) {
		out_log(LOG_ERROR, _T("glGenTextures: %d"), gl_errno);
	}
	glBindTexture(GL_TEXTURE_2D, mix_texture_name);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST + config.gl_filter_type);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST + config.gl_filter_type);
	glTexParameteri(GL_TEXTURE_2D , GL_TEXTURE_WRAP_S, GL_CLAMP);	// no wrap texture
	glTexParameteri(GL_TEXTURE_2D , GL_TEXTURE_WRAP_T, GL_CLAMP);
//	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
//	if ((gl_errno = glGetError()) != GL_NO_ERROR) {
//		out_log(LOG_ERROR, "glTexImage2D: %d", gl_errno);
//	}
	glBindTexture(GL_TEXTURE_2D, 0);	// unbound
}

void EMU::change_opengl_attr()
{
	if (!use_opengl) {
		return;
	}

	if (glIsTexture(mix_texture_name) == GL_TRUE) {
		glBindTexture(GL_TEXTURE_2D, mix_texture_name);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST + config.gl_filter_type);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST + config.gl_filter_type);
		glBindTexture(GL_TEXTURE_2D, 0);	// unbound
	}
}

void EMU::release_opengl()
{
	if (!use_opengl) {
		return;
	}

	if (glIsTexture(mix_texture_name) == GL_TRUE) {
		// release texture
		glDeleteTextures(1, &mix_texture_name);
	}
}

void EMU::set_opengl_attr()
{
	// set OpenGL atributes
#if defined(_RGB555)
	SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_ALPHA_SIZE, 0 );
	SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 0 );
	SDL_GL_SetAttribute( SDL_GL_BUFFER_SIZE, 16 );
#elif defined(_RGB565)
	SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 6 );
	SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 5 );
	SDL_GL_SetAttribute( SDL_GL_ALPHA_SIZE, 0 );
	SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 0 );
	SDL_GL_SetAttribute( SDL_GL_BUFFER_SIZE, 16 );
#elif defined(_RGB888)
	SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 8 );
	SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 8 );
	SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 8 );
	SDL_GL_SetAttribute( SDL_GL_ALPHA_SIZE, 8 );
	SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 0 );
	SDL_GL_SetAttribute( SDL_GL_BUFFER_SIZE, 32 );
#endif
	SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
//		SDL_GL_SetAttribute( SDL_GL_ACCELERATED_VISUAL, 0 );
#ifdef __WIN32__
	// if windows, multisample must be spcify
	SDL_GL_SetAttribute( SDL_GL_MULTISAMPLEBUFFERS, 0 );
	SDL_GL_SetAttribute( SDL_GL_MULTISAMPLESAMPLES, 1 );
#endif
	switch (use_opengl) {
		case 1:
			SDL_GL_SetAttribute( SDL_GL_SWAP_CONTROL, 1 );
			break;
		case 2:
			SDL_GL_SetAttribute( SDL_GL_SWAP_CONTROL, 0 );
			break;
	}
}

void EMU::set_opengl_poly(int width, int height)
{
#ifdef _DEBUG_LOG
	int v;
	SDL_GL_GetAttribute(SDL_GL_RED_SIZE, &v); out_debug(_T("SDL_GL_RED_SIZE:%d"), v);
	SDL_GL_GetAttribute(SDL_GL_GREEN_SIZE, &v); out_debug(_T("SDL_GL_GREEN_SIZE:%d"), v);
	SDL_GL_GetAttribute(SDL_GL_BLUE_SIZE, &v); out_debug(_T("SDL_GL_BLUE_SIZE:%d"), v);
	SDL_GL_GetAttribute(SDL_GL_ALPHA_SIZE, &v); out_debug(_T("SDL_GL_ALPHA_SIZE:%d"), v);
	SDL_GL_GetAttribute(SDL_GL_DEPTH_SIZE, &v); out_debug(_T("SDL_GL_DEPTH_SIZE:%d"), v);
	SDL_GL_GetAttribute(SDL_GL_BUFFER_SIZE, &v); out_debug(_T("SDL_GL_BUFFER_SIZE:%d"), v);
	SDL_GL_GetAttribute(SDL_GL_DOUBLEBUFFER, &v); out_debug(_T("SDL_GL_DOUBLEBUFFER:%d"), v);
	SDL_GL_GetAttribute(SDL_GL_MULTISAMPLEBUFFERS, &v); out_debug(_T("SDL_GL_MULTISAMPLEBUFFERS:%d"), v);
	SDL_GL_GetAttribute(SDL_GL_MULTISAMPLESAMPLES, &v); out_debug(_T("SDL_GL_MULTISAMPLESAMPLES:%d"), v);
	SDL_GL_GetAttribute(SDL_GL_SWAP_CONTROL, &v); out_log(LOG_DEBUG, _T("SDL_GL_SWAP_CONTROL:%d"), v);
#endif

	// shading
	glShadeModel( GL_SMOOTH );
	// culling
	glFrontFace( GL_CCW );
	glEnable( GL_CULL_FACE );
	glCullFace( GL_BACK );
	// clear bg color
	glClearColor( 0.0, 0.0, 0.0, 0.0 );

    // Setup our viewport : same as window size
    glViewport( 0, 0, width, height );

	glMatrixMode( GL_PROJECTION );
    glLoadIdentity();

	// change axis unit
	glOrtho(0.0, (GLdouble)width, (GLdouble)height, 0.0, -1.0, 1.0);

}

void EMU::set_pixel_format()
{
	if (pixel_format.BytesPerPixel == 4) {
		// OpenGL RGBA masks
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
		pixel_format.Rmask = 0x000000FF;
		pixel_format.Gmask = 0x0000FF00;
		pixel_format.Bmask = 0x00FF0000;
		pixel_format.Amask = 0xFF000000;
		pixel_format.Rshift = 0;
		pixel_format.Gshift = 8;
		pixel_format.Bshift = 16;
		pixel_format.Ashift = 24;
#else
		pixel_format.Rmask = 0xFF000000;
		pixel_format.Gmask = 0x00FF0000;
		pixel_format.Bmask = 0x0000FF00;
		pixel_format.Amask = 0x000000FF;
		pixel_format.Rshift = 24;
		pixel_format.Gshift = 16;
		pixel_format.Bshift = 8;
		pixel_format.Ashift = 0;
#endif
	} else if (pixel_format.BytesPerPixel == 2) {
		// OpenGL RGB masks ?
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
		pixel_format.Rmask = 0x0000001F;
		pixel_format.Gmask = 0x000007E0;
		pixel_format.Bmask = 0x0000F800;
		pixel_format.Amask = 0x00000000;
		pixel_format.Rshift = 0;
		pixel_format.Gshift = 5;
		pixel_format.Bshift = 11;
		pixel_format.Ashift = 0;
#else
		pixel_format.Rmask = 0x0000F800;
		pixel_format.Gmask = 0x000007E0;
		pixel_format.Bmask = 0x0000001F;
		pixel_format.Amask = 0x00000000;
		pixel_format.Rshift = 11;
		pixel_format.Gshift = 5;
		pixel_format.Bshift = 0;
		pixel_format.Ashift = 0;
#endif
	}
}

#endif // USE_OPENGL

