/** @file mc6843.h

	HITACHI BASIC MASTER LEVEL3 Mark5 Emulator 'EmuB-6892'
	Skelton for retropc emulator

	@par Origin
	mess & mb8877.c
	@author Sasaji
	@date   2011.11.05 -

	@brief [ fdc modoki (mc6843) ]
*/

#ifndef _MC6843_H_
#define _MC6843_H_

#include "../vm.h"
#include "../../emu.h"
#include "../device.h"

#define MC6843_MAX_EVENT 4

#define SIG_MC6843_UPDATESTATUS	0

#define MC6843_PARSE_BUFFER	8

/**
	@brief fdc modoki (mc6843) - Floppy Disk Controller
*/
class MC6843 : public DEVICE
{
private:
	FLOPPY *d_fdd;

	// registers
//	uint8 dor;        // reg0   w: disk write operation
	uint8 dir;        // reg0 r  : disk read operation
	uint8 ctar;       // reg1 r/w: current track (7bit)
	uint8 cmr;        // reg2   w: command
	uint8 isr;        // reg2 r  : interrupt status
	uint8 sur;        // reg3   w: set-up
	uint8 stra;       // reg3 r  : status a
	uint8 strb;       // reg4 r  : status b
	uint8 sar;        // reg4   w: sector address (5bit)
	uint8 gcr;        // reg5   w: general count (track  or  sector) (7bit)
	uint8 ccr;        // reg6   w: crc control (2bit)
	uint8 ltar;       // reg7   w: logical address track (=track destination) (7bit)

	// event
	int register_id[MC6843_MAX_EVENT];

	// output signals
	outputs_t outputs_irq;

	//
	int data_idx;    // current read/write position in data

	int stepcnt;	// step count
	bool now_seek;		// now seeking?
	bool now_search;	// now searching?
	bool head_load;		// now head load

	// for parse format
	uint8 parse_clk_buf[MC6843_PARSE_BUFFER];
	uint8 parse_dat_buf[MC6843_PARSE_BUFFER];
	uint8 parse_dat;
	int   parse_idx;
	int   ffw_phase;

	//for resume
#pragma pack(1)
	struct vm_state_st {
		int register_id[MC6843_MAX_EVENT];

		uint8 dir;
		uint8 ctar;
		uint8 cmr;
		uint8 isr;
		uint8 sur;
		uint8 stra;
		uint8 strb;
		uint8 sar;
		uint8 gcr;
		uint8 ccr;
		uint8 ltar;

		uint8 now;

		int data_idx;
		int stepcnt;

		char reserved[12];
	};
#pragma pack()

	void cancel_my_event(int);
	void register_my_event(int, int);

	void status_update();
	void cmd_end();

	void cmd_STZ();
	void cmd_SEK();
	void cmd_SSR();
	void cmd_SSW();
	void cmd_RCR();
	void cmd_SWD();
	void cmd_FFR();
	void cmd_FFW();
	void cmd_MSR();
	void cmd_MSW();

	int  set_delay(uint8);

	void event_seek(int);
	void event_search(int);
	void event_search2(int);
	void event_multi(int);
	void event_lost(int);

	uint8 read_data_reg();
	void write_data_reg(uint8);

	void update_stra();

	void search_track();
	void search_sector(int);

	// irq
	void set_irq(bool val);

	// for free format write
	void parse_ibm3740_format(uint8 data);
	void write_ibm3740_format();

public:
	MC6843(VM* parent_vm, EMU* parent_emu, const char* identifier) : DEVICE(parent_vm, parent_emu, identifier) {
		set_class_name("MC6843");
		init_output_signals(&outputs_irq);
	}
	~MC6843() {}

	// common functions
	void initialize();
	void reset();
	void release();
	void write_io8(uint32 addr, uint32 data);
	uint32 read_io8(uint32 addr);

	void write_signal(int id, uint32 data, uint32 mask);

	// unique function
	void set_context_irq(DEVICE* device, int id, uint32 mask) {
		register_output_signal(&outputs_irq, device, id, mask);
	}
	void set_context_fdd(FLOPPY* device) {
		d_fdd = device;
	}

	// event callback
	void event_frame();
	void event_callback(int event_id, int err);

	void save_state(FILEIO *fp);
	void load_state(FILEIO *fp);
};

#endif
